import unittest as unt
import logging
import numpy as np

from chronoclust.clustering.hddstream import HDDStream


class HelperObjTest(unt.TestCase):

    def test_set_dataset_dependent_parameters(self):
        """
        To see if the parameters are set according to the dataset size
        """
        config = {
            "beta": 0.5,
            "delta": 0.3,
            "epsilon": 10,
            "lambda": 1,
            "k": 40,
            "mu": 0.1,
            "pi": 0,
            "omicron": 0.001,
            "upsilon": 3
        }
        logger = logging.getLogger()
        hddstream = HDDStream(config, logger)
        input_data = np.random.rand(10, 2)
        hddstream.online_microcluster_maintenance(input_data, 0)

        self.assertEqual(2, hddstream.pi)
        self.assertEqual(1, hddstream.mu)
        self.assertEqual(0, hddstream.omicron)  # 0 as this should be based on previous day dataset, which is nothing.
        self.assertEqual(30, hddstream.upsilon)

        input_data = np.random.rand(30, 2)
        hddstream.online_microcluster_maintenance(input_data, 1)

        self.assertEqual(2, hddstream.pi)
        self.assertEqual(3, hddstream.mu)
        self.assertEqual(0.01, hddstream.omicron)  # this should be based on previous day dataset, which has 10 points.
        self.assertEqual(30, hddstream.upsilon)

    def test_preferred_dimension(self):
        """
        The idea in this test is to see when pi is not set to maximum, if:
        1) When the addition of a point causes variance to rise beyond delta, the dimension is no longer preferred.
        2) As soon as the number of preferred dimension is less than or equal to pi, the outlier MC is upgraded.

        The test case is deliberately set up in such as a way when the first 4 data points are added, the variance
        of each dimension is still under 0.3.
        However when the last data point is added, the last dimension is no longer preferred.
        """
        config = {
            "beta": 0.5,
            "delta": 0.3,
            "epsilon": 10,
            "lambda": 1,
            "k": 40,
            "mu": 1,
            "pi": 2,
            "omicron": 1,
            "upsilon": 1
        }

        logger = logging.getLogger()
        hddstream = HDDStream(config, logger)
        input_data = np.array([
            [0.966970507, 0.185628831, 0.861853663],
            [0.557335192, 0.324320201, 0.495929691],
            [0.698145385, 0.222617485, 0.83843284],
            [0.466592479, 0.557335192, 0.993609292]
        ])
        hddstream.online_microcluster_maintenance(input_data, 0)

        # The outlier should still have all dimensions preferred, making it not eligible to be a pcore
        self.assertEqual(0, len(hddstream.pcore_MC))
        self.assertEqual(1, len(hddstream.outlier_MC))
        pref_dim = hddstream.outlier_MC[0].preferred_dimension_vector.tolist()
        self.assertListEqual([40, 40, 40], pref_dim)

        # Now insert the last data point that will cause the last dimension to no longer be preferred.
        input_data = np.array([[0.91259336, 0.16408931, 0.06039347]])
        hddstream.online_microcluster_maintenance(input_data, 0)
        # The outlier should only have 2 dimensions preferred, making it upgraded to pcore
        self.assertEqual(1, len(hddstream.pcore_MC))
        self.assertEqual(0, len(hddstream.outlier_MC))
        pref_dim = hddstream.pcore_MC[0].preferred_dimension_vector.tolist()
        self.assertListEqual([40, 40, 1], pref_dim)

    def test_preferred_dimension_2(self):
        """
        The idea in this test is to see when decay is applied, the variance doesn't change, and pcore will remain pcore
        instead of downgraded to outlier.
        """
        config = {
            "beta": 0.001,
            "delta": 0.2,
            "epsilon": 10,
            "lambda": 0.5,
            "k": 40,
            "mu": 1,
            "pi": 3,
            "omicron": 1,
            "upsilon": 1
        }
        logger = logging.getLogger()
        hddstream = HDDStream(config, logger)
        input_data = np.array([
            [0.966970507, 0.185628831, 0.861853663],
            [0.557335192, 0.324320201, 0.495929691]
        ])
        hddstream.online_microcluster_maintenance(input_data, 0)
        pref_dim_before_decay = hddstream.pcore_MC[0].preferred_dimension_vector.tolist()

        hddstream._decay_clusters_weight(1)

        # After the decay, the pcore should still remain pcore even if it loses one preferred dimension
        self.assertEqual(1, len(hddstream.pcore_MC))
        # The preferred dimension should not change!
        pref_dim = hddstream.pcore_MC[0].preferred_dimension_vector.tolist()
        self.assertListEqual(pref_dim_before_decay, pref_dim)






"""
Chronoclust

Time series clustering algorithm.
"""

# standard import
import pandas as pd
import csv
import numpy as np
import logging
import os
import pickle

from decimal import Decimal, ROUND_HALF_UP
from collections import defaultdict

# chronoclust specific packages import
from chronoclust.clustering.hddstream import HDDStream
from chronoclust.tracking.cluster_tracker import TrackByHistoricalAssociation
from chronoclust.tracking.cluster_tracker import TrackByLineage
from chronoclust.objects.cluster import Cluster
from chronoclust.objects.microcluster import Microcluster  # for comment only
from chronoclust.scaling.scaler import Scaler

# Make this global so other function can see them as well for saving and loading
HDDSTREAM_OBJ = 'hddstream'
TRACKER_HISTORICAL_ASSOC = 'tracking_by_historical_association'
TRACKER_LINEAGE = 'tracking_by_lineage'


def run(data, output_directory, gating_centroid_file=None, normalise_data=True, restore_program=False,
        param_beta=0.2, param_delta=0.05, param_epsilon=0.03, param_lambda=2, param_k=15,
        param_mu=0.005, param_pi=0, param_omicron=0.00001, param_upsilon=2):
    """
    Run Chronoclust.

    Parameters
    ----------
    data : list
        List of names of data file for each time point.
        Make sure that they are order in order of time point (Day 1,2, etc.).
    output_directory : str
        Name of directory where the result will be stored.
        ChronoClust automatically write out results in csv format.
    gating_centroid_file : str, optional
        File containing the centroid of the data - assuming it's already gated/labelled.
        If you have pre-labelled data, you can write up the centroid of each group/population in each time point in a
        csv file and pass it here.
        It will be used to assign each Chronoclust's cluster a label that is in this file.
        The assignment is based on cluster's proximity to the groups i.e. each cluster is assigned the label that lies
        the closest to it using "weighted" euclidean distance.
    normalise_data : boolean, optional
        Whether to normalise all your data to within 0-1 range.
        What this does is call a normaliser that collate all data files and altogether normalise them.
        Note: if you call this, the final clustering result will be de-normalise to your original data range.
    restore_program : boolean, optional
        Whether to restore previous Chronoclust saved state. If this is set to true, make sure you have the images
        stored in the output_directory under program_images directory.
    param_beta : float, optional
        Value for beta parameter for Chronoclust.
    param_delta : float, optional
        Value for delta parameter for Chronoclust.
    param_epsilon : float, optional
        Value for epsilon parameter for Chronoclust.
    param_lambda : float, optional
        Value for lambda parameter for Chronoclust.
    param_k : float, optional
        Value for k parameter for Chronoclust.
    param_mu : float, optional
        Value for mu parameter for Chronoclust.
    param_pi : float, optional
        Value for pi parameter for Chronoclust.
    param_omicron : float, optional
        Value for omicron parameter for Chronoclust.
    param_upsilon : float, optional
        Value for upsilon parameter for Chronoclust.

    Returns
    -------
    None :
        Nothing is returned. This method just run Chronoclust and write up the result in output_directory.

    """

    # if we're continuing execution, the program images must be within the output directory
    program_state_dir = '{}/program_images'.format(output_directory)

    # setup logger object
    log_dir = '{}/logs'.format(output_directory)
    logger = setup_logger(log_dir)
    logger.info("Chronoclust start")

    # Get hddstream config
    config = {
        "beta": param_beta,
        "delta": param_delta,
        "epsilon": param_epsilon,
        "lambda": param_lambda,
        "k": param_k,
        "mu": param_mu,
        "pi": param_pi,
        "omicron": param_omicron,
        "upsilon": param_upsilon
    }

    # If there is a program image given, we'll search for hddstream's steam and continue from it.
    # Otherwise we'll reinitialise hddstream based on the config
    program_state_dir_exists = os.path.exists(program_state_dir)
    if restore_program and program_state_dir_exists:
        logger.info("Restoring Chronoclust state saved in {}".format(program_state_dir))
        hddstream, tracker_by_association, tracker_by_lineage = restore_program_state(program_state_dir)
        hddstream.set_logger(logger)
        hddstream.set_config(config)
    else:
        # this means either we're not restoring or the images directory does not exists
        if restore_program and not program_state_dir_exists:
            logger.warning("Restoring previous Chronoclust state not possible as program_images is not in {}".format(
                output_directory))
        logger.info("Setup new Chronoclust state")
        hddstream = HDDStream(config, logger)
        # Setup the trackers
        tracker_by_association = TrackByHistoricalAssociation()
        tracker_by_lineage = TrackByLineage()

    # find the features of the dataset by reading in the very first data file
    dataset_attributes = get_dataset_attributes(data[0])

    result_filename = f'{output_directory}/result.csv'

    result_file_header = ['timepoint', 'cumulative_size', 'pcore_ids', 'pref_dimensions'] + dataset_attributes + \
                         ['tracking_by_lineage', 'tracking_by_association']

    # Setup gating data
    gating_df = None if gating_centroid_file is None else pd.read_csv(gating_centroid_file)
    gating = defaultdict(dict)
    if gating_df is not None:

        # add predicted label to result file header
        result_file_header.append('predicted_label')

        for idx, gate in gating_df.iterrows():
            centroid = tuple(gate[dataset_attributes].values)
            pop_name = gate['PopName']
            time_point = int(gate['Day'])
            gating[time_point][centroid] = pop_name

    # write the result file header
    write_file_header(result_filename, result_file_header)

    # Setup scaler if choose to normalise data
    scaler = None
    if normalise_data:
        logger.info("Setting up scaler")
        scaler = Scaler(data)

    for timepoint, data_file in enumerate(data):
        # timepoint assuming the data_file is in order of time points

        # This is to find out the last time point processed by hddstream in previous state.
        # If it was restored, this will skip the time points that have been processed.
        # Otherwise hddstream.last_data_timestamp will be initialise to 0 and the continue won't happen.
        # Need to have the first condition as well because otherwise
        # it'll skip the very first time point if not restoring.
        if restore_program and program_state_dir_exists and hddstream.last_data_timestamp >= timepoint:
            continue

        # Read dataset
        logger.info("Processing dataset {}".format(timepoint))
        dataset = pd.read_csv(data_file, header=0, sep=',').to_numpy()

        # normalise the dataset if needed
        if normalise_data:
            logger.info("Scaling dataset {}".format(timepoint))
            dataset = scaler.scale_data(dataset)

        # Start clustering
        hddstream.online_microcluster_maintenance(dataset, timepoint)
        hddstream_pcore_id_to_object_dict = {x.id[0]: x for x in hddstream.pcore_MC}

        for cluster in hddstream.final_clusters:

            # Round the cluster weight to 1 decimal place
            rounded_weight = Decimal(str(cluster.cumulative_weight)).quantize(Decimal('1.1'), rounding=ROUND_HALF_UP)

            cluster = Cluster(list(cluster.id), cluster.cluster_centroids, rounded_weight,
                                  cluster.preferred_dimension_vector)
            cluster.add_pcore_objects(hddstream_pcore_id_to_object_dict)

            tracker_by_lineage.add_new_child_cluster(cluster)

        # Start tracking by lineage
        tracker_by_lineage.calculate_ids()

        # Start tracking by historical associates
        tracker_by_association.set_current_clusters(tracker_by_lineage.child_clusters)
        tracker_by_association.track_cluster_history()

        # Start writing out result so we don't lose any result if program crashes.
        # Starting with overall result file
        write_result_file(gating, result_filename, timepoint, tracker_by_association, scaler=scaler)

        cluster_points_filename = f'{output_directory}/cluster_points_D{timepoint}.csv'
        clusters = tracker_by_lineage.child_clusters
        pcoreMCs = hddstream.pcore_MC
        outlierMCs = hddstream.outlier_MC
        write_datapoints_details(dataset_attributes, clusters, pcoreMCs, outlierMCs, logger, cluster_points_filename,
                                 True, scaler)

        # Prepare for the next time point
        tracker_by_lineage.transfer_child_to_parent()
        tracker_by_association.transfer_current_to_previous()

        # Save program state
        logger.info("Saving Chronoclust state for timepoint {}".format(timepoint))
        save_program_state(hddstream, output_directory, tracker_by_association, tracker_by_lineage)

    # Write out the hddstream setting
    settings_filename = f'{output_directory}/parameters.csv'
    with open(settings_filename, 'w') as f:
        w = csv.DictWriter(f, config.keys())
        w.writeheader()
        w.writerow(config)

    # Log finish point
    logger.info('Chronoclust finish')


def write_result_file(gating, result_filename, timepoint, tracker_by_association, scaler):
    result = []
    gating_now = gating.get(timepoint)
    for cluster in tracker_by_association.current_clusters:
        historical_assoc_as_str = cluster.get_historical_associates_as_str()
        pcore_ids_as_str = cluster.get_pcore_ids_as_str()
        pref_dims_as_str = cluster.get_preferred_dimensions_as_str()

        # Combine all the result components into a single array
        array_rep = [timepoint, cluster.cumulative_weight, pcore_ids_as_str, pref_dims_as_str]

        # write out cluster centroid. If scaler is given, it means the data was normalised by chronoclust.
        # Hence the result must be denormalised.
        if scaler:
            centroid = scaler.reverse_scaling([cluster.centroid]).tolist()[0]
            centroid = np.round(centroid, 5).tolist()
        else:
            centroid = np.round(cluster.centroid, 5).tolist()

        array_rep.extend(centroid)
        array_rep.append(cluster.id)
        array_rep.append(historical_assoc_as_str)

        # It means we have gating information
        if bool(gating_now):
            # closest_gate_projected is the closest gate calculation based on projected distance.
            # pick one and add it to the array_rep.
            closest_gate_projected = find_closest_gating(gating_now, cluster, scaler)
            array_rep.append(closest_gate_projected)

        result.append(array_rep)
    append_to_file(result_filename, result)


def write_datapoints_details(dataset_attributes, clusters, pcoreMCs, outlierMCs, logger, cluster_points_filename,
                             sort_output, scaler):
    """
    Write out the information of the datapoints that has just been clustered into a csv file.

    Parameters
    ----------
    dataset_attributes : list of str
        The features of the dataset. Needed to write out the header of the csv file as well as separating the value
        of the features in the datapoints
    clusters : list of Cluster
        List of Cluster object. This is the list of clusters produced by HDDStream.
    pcoreMCs : list of Microcluster
        List of PCore Microcluster
    outlierMCs : list of Microcluster
        List of outlier Microcluster
    logger : Logger
        Logger object from logging module
    cluster_points_filename : str
        The name of the csv file for output
    sort_output : bool
        Whether to sort the output datapoints based on its id, i.e. the order in which the input datapoints is presented
    scaler : Scaler
        Scaler object. If given i.e. not none, then each datapoint will be denormalised using the scaler. Otherwise no

    Returns
    -------
    None

    """

    # Id is the order in which the points are read in (line by line, by original order of the input dataset)
    # the ordering could be useful to some work flow that requires clustered data points to be written out in
    # the same order as it comes in
    write_file_header(cluster_points_filename, ['id', 'cluster_id'] + dataset_attributes)
    # This will extract all the points that are clustered
    # The following is used to keep track of pcore that is part of cluster. So we can print out the ones that
    # were not part of any cluster
    clustered_pcore_id = set()
    clustered_dp = defaultdict(list)
    for cluster in clusters:
        cluster_id = cluster.id
        for pcore_mc in cluster.pcore_objects:
            # This will happen if there are no points belonging to current day get clustered into
            # one of the pcore that's part of current day cluster.
            # If we don't have the try condition below, the scaler will throw an error.
            try:
                points = list(pcore_mc.points.values())
                if scaler:
                    points = scaler.reverse_scaling(points).tolist()

                points_id = list(pcore_mc.points.keys())
            except ValueError:
                logger.info("WARN: Pcore {} did not receive new datapoints.".format(pcore_mc.id))
                continue

            # zip is ok here as both array MUST be of same length i.e. each point has an id
            for point, point_id in zip(points, points_id):
                clustered_dp['id'].append(point_id)
                clustered_dp['cluster_id'].append(cluster_id)
                clustered_dp['values'].append([p for p in point])

            clustered_pcore_id.add(pcore_mc.id[0])  # the id is list type. Ok stupid but TODO to fix it
    # This will extract all the points that are in the pcore-MC but NOT in a cluster reported at the end.
    unclustered_pcoreMCid = set([p_mc.id[0] for p_mc in pcoreMCs]) - clustered_pcore_id
    unclustered_pcoreMC = [p_mc for p_mc in pcoreMCs if p_mc.id[0] in unclustered_pcoreMCid]
    unclustered_pcore_dp = obtain_unclustered_dp_info(unclustered_pcoreMC, scaler)

    # This will extract all the points that are in outlier. We'll label them as noise.
    unclustered_outlier_dp = obtain_unclustered_dp_info(outlierMCs, scaler)

    # join all the data points detail together in one big dictionary
    all_dp_dicts = [clustered_dp, unclustered_pcore_dp, unclustered_outlier_dp]
    dp_details = defaultdict(list)
    for a_dp_dict in all_dp_dicts:
        for k, v in a_dp_dict.items():
            dp_details[k].extend(v)

    # now we need to separate each feature's value into different key in the dictionary
    # atm they're all grouped together under a key call "values"
    # dp_details['values'] contains [[1,2,3],[4,5,6]] where 1 and 4 are for feature A, [1,2,3] is for datapoint X
    for dp_feature_vals in dp_details['values']:
        for feature_name, dp_feature_val in zip(dataset_attributes, dp_feature_vals):
            dp_details[feature_name].append(dp_feature_val)

    # then we need to remove the key
    dp_details.pop('values')

    # now we create dataframe for it
    dp_details_df = pd.DataFrame(dp_details)

    # If the option to sort cluster points is passed, then we need to sort the points before it's printed
    # extra condition because you can't sort if there is no data points
    if sort_output:
        dp_details_df.sort_values(by=['id'], inplace=True)

    # write out the datapoints details
    dp_details_df.to_csv(cluster_points_filename, index=False)


def obtain_unclustered_dp_info(mcs, scaler):
    """
    Extract any unclustered datapoints information from microcluster.
    For each datapoint, the method extract the id as well as the feature values.
    It will then put those into a dictionary of list with key id for id, values for feature values,
    as well as add "none" as cluster_id as these are unclustered data points.

    Parameters
    ----------
    mcs : list of Microcluster
        List of Microcluster which datapoints need to be extracted
    scaler : Scaler
        Scaler object. If given, the datapoint will be denormalised using the scaler

    Returns
    -------
    defaultdict
        A dictionary containing the details of each datapoint within mcs

    """
    dp_details = defaultdict(list)
    for mc in mcs:
        try:
            points = list(mc.points.values())
            if scaler:
                points = scaler.reverse_scaling(points).tolist()
            pts_id = list(mc.points.keys())
        except ValueError:
            continue

        # zip is ok here as both array MUST be of same length i.e. each point has an id
        for pt, pt_id in zip(points, pts_id):
            dp_details['id'].append(pt_id)
            dp_details['cluster_id'].append("None")
            dp_details['values'].append([p for p in pt])

    return dp_details


def save_program_state(hddstream, output_dir, tracker_by_association, tracker_by_lineage):
    """
    Save the states so it can carry on where it left off when restarted. Save it in the folder in output
    Note only saving the very last state. This is to allow us to just resume from the very last checkpoint.
    Improvement MAYBE is to categorise and save all checkpoints

    Parameters
    ----------
    hddstream : HDDStream
        HDDStream object which state need to be saved
    output_dir : str
        Directory where the image (state) will be stored (under subfolder program_images)
    tracker_by_association : TrackByHistoricalAssociation
        TrackByHistoricalAssociation object which state need to be saved
    tracker_by_lineage : TrackByLineage
        TrackByLineage object which state need to be saved

    Returns
    -------
    None

    """

    program_state_dir_for_saving = "{}/program_images".format(output_dir)
    if not os.path.exists(program_state_dir_for_saving):
        os.mkdir(program_state_dir_for_saving)
    with open('{}/{}'.format(program_state_dir_for_saving, HDDSTREAM_OBJ), 'wb') as f:
        pickle.dump(hddstream, f)
    with open('{}/{}'.format(program_state_dir_for_saving, TRACKER_HISTORICAL_ASSOC), 'wb') as f:
        pickle.dump(tracker_by_association, f)
    with open('{}/{}'.format(program_state_dir_for_saving, TRACKER_LINEAGE), 'wb') as f:
        pickle.dump(tracker_by_lineage, f)


def restore_program_state(program_state_dir):
    """
    Restore program's state (hddstream and tracker) using pickle.

    Parameters
    ----------
    program_state_dir : str
        Directory where the image is stored

    Returns
    -------
    HDDStream
        HDDStream object restored from the image
    TrackByHistoricalAssociation
        TrackByHistoricalAssociation object restored from the image
    TrackByLineage
        TrackByLineage object restored from the image

    """

    with open('{}/{}'.format(program_state_dir, HDDSTREAM_OBJ), 'rb') as f:
        hddstream = pickle.load(f)

    with open('{}/{}'.format(program_state_dir, TRACKER_HISTORICAL_ASSOC), 'rb') as f:
        tracker_by_association = pickle.load(f)

    with open('{}/{}'.format(program_state_dir, TRACKER_LINEAGE), 'rb') as f:
        tracker_by_lineage = pickle.load(f)

    return hddstream, tracker_by_association, tracker_by_lineage


def setup_logger(log_dir):
    # create directory of the log file if it doesn't exist
    if not os.path.exists(log_dir):
        os.makedirs(log_dir)

    # initialise log file
    logger_filename = '{}/Chronoclust.log'.format(log_dir)

    logging.basicConfig(filename=logger_filename, format='%(asctime)s [%(levelname)-8s] %(message)s')
    logger = logging.getLogger()
    logger.setLevel(logging.INFO)
    return logger


def write_file_header(filename, header):
    with open(filename, 'w') as f:
        csv.writer(f).writerow(header)


def append_to_file(filename, content):
    with open(filename, 'a') as f:
        csv.writer(f).writerows(content)


def get_dataset_attributes(dataset_file):
    # let pandas infer whether the dataset file is compressed or not based on the filename extension.
    return pd.read_csv(dataset_file, sep=',', header=None).iloc[0].values.tolist()


def find_closest_gating(gating_dict, cluster, scaler):
    closest_distance_projected = None
    closest_gate_projected = None

    for centroid, label in gating_dict.items():
        if scaler:
            centroid_norm = scaler.scale_data([centroid])[0].tolist()
        else:
            centroid_norm = centroid

        dist_projected = cluster.get_projected_dist_to_point(np.array(centroid_norm))
        if closest_distance_projected is None or dist_projected < closest_distance_projected:
            closest_distance_projected = dist_projected
            closest_gate_projected = label

    return closest_gate_projected



"""
Scaler module to normalise data to range of 0 and 1.
The scaler uses Scikit Learn's MinMaxScaler.
Scikit-learn: Machine Learning in Python, Pedregosa et al., JMLR 12, pp. 2825-2830, 2011.
"""
import pandas as pd

from sklearn.preprocessing import MinMaxScaler


class Scaler(object):

    def __init__(self, data_files=None):
        """
        Initialise scaler object. If no data_files is given, then nothing is done.
        TODO: make it give out an error message if no data_files is given.

        Parameters
        ----------
        data_files : list
            List of str containing the filenames of the time-series data (1 data file per time point).
        """

        self.scaler = MinMaxScaler()
        self.input_data = []

        if data_files is not None:
            all_data_points = []
            for filename in data_files:
                # read in all the data files, convert to numpy list.
                df = pd.read_csv(filename, header=0, sep=',').to_numpy()
                # add it to the big list containing all data points.
                all_data_points.extend(df)

            # fit the scaler
            self.fit_scaler(all_data_points)

    def fit_scaler(self, data):
        self.scaler.fit(data)
        # when fitting new scaler, you update the input data used to "setup" the scaler
        self.set_input_data(data)

    def scale_data(self, data):
        return self.scaler.transform(data)

    def reverse_scaling(self, data):
        return self.scaler.inverse_transform(data)

    def set_input_data(self, data):
        self.input_data = data

    def get_input_data(self):
        return self.input_data




import unittest as unt
import pandas as pd
import os
import shutil

from chronoclust.app import run

# note change this when refactoring filename
current_script_dir = os.path.realpath(__file__).split('/{}'.format(os.path.basename(__file__)))[0]
out_dir = '{}/test_files/output'.format(current_script_dir)


class IntegrationTestNoCluster(unt.TestCase):
    @classmethod
    def setUpClass(cls):
        """
        Run chronoclust

        Returns
        -------
        None
        """

        data = ['{codedir}/test_files/dataset/subset_dataset/synthetic_d{t}.csv.gz'.format(
            codedir=current_script_dir,
            t=day
        ) for day in range(5)]

        config = {
            "beta": 1.0,
            "delta": 0.05,
            "epsilon": 0.03,
            "lambda": 2,
            "k": 4,
            "mu": 1.0,
            "pi": 3,
            "omicron": 0.000000435,
            "upsilon": 6.5
        }

        run(data=data, output_directory=out_dir, param_beta=config['beta'], param_delta=config['delta'],
            param_epsilon=config['epsilon'], param_lambda=config['lambda'], param_k=config['k'], param_mu=config['mu'],
            param_pi=config['pi'], param_omicron=config['omicron'], param_upsilon=config['upsilon'])

    @classmethod
    def tearDownClass(cls):
        """
        Remove the output folder

        Returns
        -------
        None
        """

        shutil.rmtree('{}/test_files/output'.format(current_script_dir))

    def test_result(self):

        """
        Normal run where there are no clusters.
        The dataset is based on synthetic dataset used for original Chronoclust paper.
        Only testing the result file here.

        Returns
        -------
        None

        """

        result_df = pd.read_csv('{}/result.csv'.format(out_dir))

        # test that we don't have any row in the result file.
        self.assertEqual(0, result_df.shape[0])

        # test the cluster points
        for i in range(5):
            cluster_dp_df = pd.read_csv('{}/cluster_points_D{}.csv'.format(out_dir, i))

            # test we only have 10 rows as there are only 10 rows in the dataset
            self.assertEqual(10, cluster_dp_df.shape[0])

            # check all datapoints are assigned to no cluster
            for row in cluster_dp_df.itertuples():
                cluster_id = getattr(row, 'cluster_id')
                self.assertEqual('None', cluster_id)


if __name__ == '__main__':
    unt.main()



#!/usr/bin/env python

"""
PreDeCon algorithm. See the following paper for more information.
[2] Bohm, Christian, et al. "Density connected clustering with local subspace preferences." Data Mining, 2004.
The paper will be referred as paper[2].
"""
import numpy as np

from chronoclust.objects.predecon_mc import PredeconMC
from chronoclust.utilities import predeconmc_functions as predfunc

__author__ = "Givanna Putri, Deeksha Singh, Mark Read, and Tao Tang"
__copyright__ = "Copyright 2017, Cytoclust Project"
__credits__ = ["Givanna Putri", "Deeksha Singh", "Mark Read", "Tao Tang"]
__version__ = "0.0.1"
__maintainer__ = "Givanna Putri"
__email__ = "ghar1821@uni.sydney.edu.au"
__status__ = "Development"


class PreDeCon(object):
    def __init__(self, datapoints, dataset_dimensionality, epsilon, delta, lambbda, mu, k):
        """
        Create the PreDeCon object. The parameters used here have the same name and meaning as Bohm paper. Please
        refer to the paper for more information.

        Args:
            datapoints ((:obj:`dict`): Datapoints to be clustered.
            dataset_dimensionality (int): Number of dimensions in the dataset.
            epsilon (float): Epsilon for PreDeCon.
            delta (float): Delta for PreDeCon.
            lambbda (float): Lambda for PreDeCon.
            mu (float): Mu for PreDeCon.
            k (int): K constant for PreDeCon.
            logger (:obj:`str`): Directory to put log file for PreDeCon execution.
        """
        self.datapoints = datapoints
        self.epsilon = epsilon
        self.epsilon_squared = epsilon ** 2
        self.dataset_dimensionality = dataset_dimensionality
        self.delta = delta
        self.delta_squared = delta ** 2
        self.lambbda = lambbda
        self.mu = mu
        self.k = k
        self.clusters = []

    def run(self):
        """
        Method to run PreDeCon.

        Returns:
            None.
        """
        self._find_weighted_neighbours()

        datapoints_items = self.datapoints.items()
        delta_squared = self.delta_squared
        k = self.k

        for datapt_id, datapoint in datapoints_items:

            if datapoint.is_unclassified():
                if datapoint.is_core():

                    # Need to do it this way as for offline clustering, the cluster id must be an empty set.
                    new_cluster_id = set()
                    datapoint_values = len(datapoint.get_centroid())

                    new_cluster = PredeconMC(cluster_CF1=np.zeros(datapoint_values),
                                               cluster_CF2=np.zeros(datapoint_values), id=new_cluster_id,
                                               centroid=np.zeros(datapoint_values),
                                               is_core_cluster=False,
                                               cluster_cumulative_weight=0)

                    # Try to expand the new_cluster with the datapoint
                    self._expand(new_cluster, datapoint)

                    new_cluster.update_preferred_dimensions(delta_squared, k)

                    # Sanity check to make sure we do not include empty cluster.
                    if new_cluster.cumulative_weight > 0:
                        self.clusters.append(new_cluster)

                else:
                    datapoint.set_noise()

    def _expand(self, cluster, datapoint):
        """
        Expand the cluster if possible. See Figure 4 line 6-14 in paper[2]

        Args:
            cluster (:obj:Microcluster): Cluster to be expanded
            datapoint (:obj:Datapoint): Datapoint to expand the cluster with.

        Returns:
            None.
        """

        # a copy of potential_cluster_members because it will be modified. We don't want to modify parameters. This
        # will make a copy as the weighted_neighbour_pts contains ids as int.
        queue = list(datapoint.weighted_neighbour_pts)
        datapoints = self.datapoints
        while len(queue) > 0:
            # q in paper[2]
            first_datapt = datapoints[queue.pop(0)]
            # R in paper[2]
            directly_reachable_pts = self._find_directly_reachable_points(first_datapt, list(datapoints.keys()))

            for dir_reachable_pt_id in directly_reachable_pts:
                # x in paper[2] figure 4.
                x = datapoints[dir_reachable_pt_id]

                # line 10-14 of figure 4 in paper[2]
                if x.is_unclassified():
                    queue.append(dir_reachable_pt_id)
                if x.is_unclassified() or x.is_noise():
                    x.set_classified()
                    x.merge_mc(cluster)

    def _set_is_core_pt(self):
        """
        Method that will set each datapoint's core status. See Definition 6 in paper[2].

        Returns:
            None.
        """
        datapoints_items = self.datapoints.items()
        lambbda = self.lambbda
        mu = self.mu

        for datapt_id, datapt in datapoints_items:
            datapt.set_is_core()(lambbda, mu)

    def _find_weighted_neighbours(self):
        """
        Method to find and set each datapoint's weighted neighbours. Done by first finding normal neighbours and
        calculating subspace preference vector for ALL datapoints.
        Then start calculating weighted neighbours based on those 2 information.

        Returns:
            None.
        """
        datapoints = self.datapoints
        datapoints_items = datapoints.items()
        epsilon_squared = self.epsilon_squared

        for datapt_id, datapt in datapoints_items:
            # Find all the neighbour points and calculate the subspace preference vector
            datapt.neighbour_pts = self._find_neighbour_points(datapt)
            datapt.subspace_preference_vector = self._calculate_subspace_preference_vector(datapt)

        # This MUST be done after the preference vectors for all points must have been calculated.
        for datapt_id, datapt in datapoints_items:
            for neighbour_pt_id in datapt.neighbour_pts:
                dist = self._calculate_general_weighted_dist_squared(datapt, datapoints[neighbour_pt_id])
                if dist <= epsilon_squared:
                    datapt.weighted_neighbour_pts.append(neighbour_pt_id)

    def _find_neighbour_points(self, point):
        """
        Find all points within point's neighbourhood.

        Args:
            point (:obj:Datapoint): Datapoint we are interested in finding the neighbourhood for

        Returns:
            list: List of neighbourhood points' id.
        """
        neighbour_points_id = []

        datapoints_items = self.datapoints.items()
        epsilon = self.epsilon

        pt_centroid = np.array(point.get_centroid(), dtype='float')

        for datapt_id, datapt in datapoints_items:
            # euclidean_dist = self._calculate_euclidean_dist(datapt.get_centroid(), point.get_centroid())
            dtpt_centroid = np.array(datapt.get_centroid(), dtype='float')

            euclidean_dist = predfunc.calculate_euclidean_dist(dtpt_centroid, pt_centroid)

            # See beginning of chapter 3 of paper[2] for neighbourhood points criteria
            if euclidean_dist <= epsilon:
                neighbour_points_id.append(datapt_id)

        return neighbour_points_id

    def _calculate_subspace_preference_vector(self, mc):
        """
        Calculate the subspace preference vector of a mc (w_p in paper[2]). Refer to definition 3 of paper[2]

        Args:
            mc (PredeconMC): PredeconMC whose subspace preference vector is to be found

        Returns:
            Array: Subspace preference vector represented by an array.
        """
        delta = self.delta
        k = self.k
        all_mcs = self.datapoints

        # get the centroid of the mc of interest and its neighbours
        mc_centroid = mc.get_centroid()
        neighbours = [all_mcs[p].get_centroid() for p in mc.neighbour_pts]

        subspace_preference_vector = []
        for dim in range(len(mc_centroid)):
            mc_centroid_dim = np.array(mc_centroid[dim])
            neighbour = np.array([n[dim] for n in neighbours])
            variance = predfunc.calculate_variance_along_dimension(mc_centroid_dim, neighbour)
            if variance <= delta:
                subspace_preference_vector.append(k)
            else:
                subspace_preference_vector.append(1)
        return subspace_preference_vector

    def _calculate_general_weighted_dist_squared(self, p, q):
        """
        Method to calculate weighted distance (squared version) between 2 points (dist_pref in paper [2]).
        See definition 4 in paper[2].

        Args:
            p (:obj:Datapoint): Point 1.
            q (:obj:Datapoint): Point 2.

        Returns:
            Float: Squared weighted distance.
        """
        p_centroid = np.array(p.get_centroid())
        q_centroid = np.array(q.get_centroid())
        p_pref_vect = np.array(p.subspace_preference_vector)
        q_pref_vect = np.array(q.subspace_preference_vector)

        dist_p_q = predfunc.calculate_weighted_dist_squared(p_pref_vect, p_centroid, q_centroid)
        dist_q_p = predfunc.calculate_weighted_dist_squared(q_pref_vect, q_centroid, p_centroid)

        return(max(dist_p_q, dist_q_p))


    def _find_directly_reachable_points(self, point, potential_directly_reachable_points):
        """
        Given a point, find all directly reachable points.

        Args:
            point (:obj:Datapoint): Point whose directly reachables points we want to find. If following Figure 4 in
                paper[2] this should be q.

        potential_directly_reachable_points (:obj:List): List of points that is potentially directly reachable from
            the point.

        Returns:
            List: List of directly reachable points
        """
        datapoints = self.datapoints
        lambbda = self.lambbda

        dir_reachable_pts = []
        for datapt_id in potential_directly_reachable_points:
            point_is_core = point.is_core()
            pdim_datapt_less_than_threshold = datapoints[datapt_id].get_pdim() <= lambbda
            datapt_is_neighbour_of_point = datapt_id in point.weighted_neighbour_pts

            if point_is_core and pdim_datapt_less_than_threshold and datapt_is_neighbour_of_point:
                dir_reachable_pts.append(datapt_id)
        return dir_reachable_pts



import string

from collections import defaultdict
from collections import deque


class TrackByLineage(object):
    def __init__(self):
        self.child_clusters = []
        self.parent_clusters = []
        self.letters = deque(string.ascii_uppercase)
        self.split_per_id = dict((el, 0) for el in self.letters)

    def repopulate_letters(self, iteration):
        letters = []
        for alphabet in string.ascii_uppercase:
            letters.append(alphabet * iteration)
        self.letters = deque(letters)

    def get_new_letter(self):
        new_letter = self.letters.popleft()
        if len(self.letters) == 0:
            self.repopulate_letters(len(new_letter) + 1)
        return new_letter

    def add_new_child_cluster(self, cluster):
        self.child_clusters.append(cluster)

    def calculate_ids(self):
        # Just for presentation purposes.
        # We add the if condition just in case there is a cluster which cumulative weight is not given.
        # It's made optional.
        if None not in [x.cumulative_weight for x in self.child_clusters]:
            self.child_clusters.sort(key=lambda x: x.cumulative_weight)

        offspring = defaultdict(list)
        # Get the cluster label from previous day.
        parent_pcores_to_id = self.get_parent_pcore_to_id()
        for cluster in self.child_clusters:
            cluster.set_parents(parent_pcores_to_id=parent_pcores_to_id)

            # Cluster only contains new pcores
            if len(cluster.parents) == 0:
                new_letter = self.get_new_letter()
                cluster.add_parent(id=new_letter)

            # Group each parent
            for parent in cluster.get_parents():
                offspring[parent].append(cluster)

        for parent, children in offspring.items():
            # Sort children descendingly by number of parent MCs they contain.
            # https://stackoverflow.com/questions/403421/how-to-sort-a-list-of-objects-based-on-an-attribute-of-the-objects
            children = sorted(children, key=lambda x: len(x.pcore_ids), reverse=True)

            # If cluster B split into B and B|1 in day 1, then split again to produce B, B|1, and B|2 in day
            # 2, i.e. B split again in day 2, we need to make sure we do not assign the new split label B|1
            # as in day 1, cluster B already split into B|1. To do this, we need to remember the number of
            # split cluster B has had.
            parent_num_split = self.split_per_id.get(parent, 0)

            for i in range(0, len(children)):
                child = children[i]
                # Cluster with biggest number of pcores. It inherits the previous day label.
                if i == 0:
                    child.add_id(parent)
                else:
                    parent_num_split += 1
                    child.add_id(f'{parent}|{parent_num_split}')
            self.split_per_id[parent] = parent_num_split

        # Assign the id to each child. Handle merges if a cluster has separate parents.
        self.assign_child_id()

    def get_parent_pcore_to_id(self):
        pcore_to_id = {}
        for parent in self.parent_clusters:
            pcores = parent.pcore_ids
            id = parent.id
            for pcore in pcores:
                pcore_to_id[pcore] = id
        return pcore_to_id

    def transfer_child_to_parent(self):
        self.parent_clusters = self.child_clusters
        self.child_clusters = []

    def assign_child_id(self):
        """
        Not a simple join method because of the merging. When there are multiple items in self.label set, that means
        there has been a merge somewhere along the line.

        When merging we need to enclose the merging clusters' label in parenthesis.
        Moreover, we need to separate the merging clusters' label by comma.

        For example, cluster A|1 and B merges. The resulting labels should be (A|1, B).
        If that cluster merges again, say with cluster B|2, then the label will become ((A|1, B), B|2).
        """
        for child in self.child_clusters:
            if len(child.id) == 1:
                child.id = ''.join(sorted(child.id))
            else:
                labels_as_list = list(sorted(child.id))
                label_with_parenthesis = ['(', labels_as_list[0], ',', labels_as_list[1], ')']
                if len(label_with_parenthesis) > 2:
                    for i in range(2, len(labels_as_list)):
                        label_with_parenthesis = ['('] + label_with_parenthesis + [',', labels_as_list[i], ')']

                child.id = ''.join(label_with_parenthesis)


class TrackByHistoricalAssociation(object):
    def __init__(self):
        self.current_clusters = []
        self.previous_timepoint_clusters = []

    def set_current_clusters(self, clusters):
        self.current_clusters = clusters

    def track_cluster_history(self):
        if len(self.previous_timepoint_clusters) == 0:
            for cluster in self.current_clusters:
                cluster.add_historical_associate(None)
            return

        # TODO need improvement. Too many loops here. Can maybe get each cluster to do the evaluation
        for cluster in self.current_clusters:

            for pcore in cluster.pcore_objects:
                closest_distance = None
                closest_previous_timepoint_cluster = None
                # This is only for finding out which pcore is the closest.
                closest_pcore = None

                for previous_timepoint_cluster in self.previous_timepoint_clusters:
                    for previous_pcore in previous_timepoint_cluster.pcore_objects:
                        projected_distance = pcore.get_projected_dist_to_point(previous_pcore.cluster_centroids)
                        if closest_distance is None or projected_distance < closest_distance:
                            closest_distance = projected_distance
                            closest_previous_timepoint_cluster = previous_timepoint_cluster.id
                            closest_pcore = previous_pcore.id

                cluster.add_historical_associate(closest_previous_timepoint_cluster)
                cluster.add_historical_associate_pcore(closest_pcore)

    def transfer_current_to_previous(self):
        self.previous_timepoint_clusters = self.current_clusters
        self.current_clusters = []



import unittest as unt
import pandas as pd
import numpy as np
import shutil
import os

from chronoclust.app import run

# note change this when refactoring filename
current_script_dir = os.path.realpath(__file__).split('/{}'.format(os.path.basename(__file__)))[0]
out_dir = '{}/test_files/output'.format(current_script_dir)


class IntegrationTestNormal(unt.TestCase):
    @classmethod
    def setUpClass(cls):
        """
        Run chronoclust

        Returns
        -------
        None
        """

        # We need to write the input xml file first. Template is already given.
        # This is because the location of the data file will vary from machine to machine, and the last thing
        # we want to do is changing the xml file as we transfer between machine.

        data = ['{codedir}/test_files/dataset/full_dataset/synthetic_d{t}.csv.gz'.format(
            codedir=current_script_dir,
            t=day
        ) for day in range(5)]

        config = {
            "beta": 0.2,
            "delta": 0.05,
            "epsilon": 0.03,
            "lambda": 2,
            "k": 4,
            "mu": 0.01,
            "pi": 3,
            "omicron": 0.000000435,
            "upsilon": 6.5
        }

        gating = '{}/test_files/dataset/full_dataset/gating_centroids.csv'.format(current_script_dir)

        run(data=data, output_directory=out_dir, gating_centroid_file=gating, param_beta=config['beta'],
            param_delta=config['delta'], param_epsilon=config['epsilon'], param_lambda=config['lambda'],
            param_k=config['k'], param_mu=config['mu'], param_pi=config['pi'], param_omicron=config['omicron'],
            param_upsilon=config['upsilon'])


    @classmethod
    def tearDownClass(cls):
        """
        Remove the output folder

        Returns
        -------
        None
        """

        shutil.rmtree('{}/test_files/output'.format(current_script_dir))

    def test_normal_result(self):

        """
        Normal run where there are clusters, all kinds of MCs, and unclustered data.
        The dataset is based on synthetic dataset used for original Chronoclust paper.
        Only testing result file
        Returns
        -------
        None

        """

        result_df = pd.read_csv('{}/result.csv'.format(out_dir))
        expected_result_df = pd.read_csv('{}/test_files/expected_output/result.csv'.format(current_script_dir))

        # test header is the same, except for one extra in the result file that is "predicted_label" as we run it
        # with gating file.
        np.testing.assert_array_equal(expected_result_df.columns, result_df.columns)

        # test same number of rows
        self.assertEqual(result_df.shape[0], expected_result_df.shape[0])

        # test the cluster ids are all the same
        np.testing.assert_equal(result_df['tracking_by_lineage'].to_numpy(), expected_result_df['tracking_by_lineage'].to_numpy())

        # test each row is the same (yes the order should not change)
        for ex_row in expected_result_df.itertuples():
            ex_cluster_id = getattr(ex_row, 'tracking_by_lineage')
            ex_timepoint = getattr(ex_row, 'timepoint')

            row = result_df[(result_df['tracking_by_lineage'] == ex_cluster_id) &
                             (result_df['timepoint'] == ex_timepoint)]

            self.assertEqual(getattr(ex_row, 'cumulative_size'), row['cumulative_size'].to_numpy()[0])
            self.assertEqual(getattr(ex_row, 'pcore_ids'), row['pcore_ids'].to_numpy()[0])
            self.assertEqual(getattr(ex_row, 'pref_dimensions'), row['pref_dimensions'].to_numpy()[0])
            np.testing.assert_approx_equal(getattr(ex_row, 'x'), row['x'].to_numpy()[0])
            np.testing.assert_approx_equal(getattr(ex_row, 'y'), row['y'].to_numpy()[0])
            np.testing.assert_approx_equal(getattr(ex_row, 'z'), row['z'].to_numpy()[0])

            # the order of multiple predecessor clusters can be arbitrary. Just need to check the content
            expected_predecessors = getattr(ex_row, 'tracking_by_association').split("&")
            actual_predecessors = row['tracking_by_association'].to_numpy()[0].split("&")
            for a in actual_predecessors:
                self.assertIn(a, expected_predecessors)

            self.assertEqual(getattr(ex_row, 'predicted_label'), row['predicted_label'].to_numpy()[0])

        # TODO move this to separate method
        # test cluster points
        for i in range(5):
            cluster_dp_df = pd.read_csv('{}/cluster_points_D{}.csv'.format(out_dir, i))
            expected_df_df = pd.read_csv(
                '{}/test_files/expected_output/cluster_points_D{}.csv'.format(current_script_dir, i))

            # test header is the same
            np.testing.assert_array_equal(cluster_dp_df.columns, expected_df_df.columns)

            # test same number of rows
            self.assertEqual(expected_df_df.shape[0], cluster_dp_df.shape[0])

            # test the cluster ids are all the same, disregard order
            cluster_ids = expected_df_df['cluster_id'].unique()
            np.testing.assert_array_equal(sorted(cluster_ids),
                                          sorted(cluster_dp_df['cluster_id'].unique()))

            # test each cluster has same member
            for cl_id in cluster_ids:

                ex_row = expected_df_df[(expected_df_df['cluster_id'] == cl_id)].sort_values(by=['x', 'y', 'z'])
                row = cluster_dp_df[(cluster_dp_df['cluster_id'] == cl_id)].sort_values(by=['x', 'y', 'z'])

                # test we have same number of datapoints in the cluster
                self.assertEqual(ex_row.shape[0], row.shape[0])

                # test the datapoints therein are the same
                # may not work well as it truly depends on how the rounding works for each data point
                ex_dps = ex_row[['x', 'y', 'z']].to_numpy()
                dps = row[['x', 'y', 'z']].to_numpy()

                for ex_dp, dp in zip(ex_dps, dps):
                    np.testing.assert_array_almost_equal(ex_dp, dp, decimal=1)

            # test order of the data points in the output file (must be the same as the input file)
            dataset = pd.read_csv(
                '{}/test_files/dataset/full_dataset/synthetic_d{}.csv.gz'.format(current_script_dir, i))
            cluster_dp_df = cluster_dp_df.round(2)
            data_rows = dataset[['x', 'y', 'z']].to_numpy()
            cluster_rows = cluster_dp_df[['x', 'y', 'z']].to_numpy()

            for d, c in zip(data_rows, cluster_rows):
                np.testing.assert_array_almost_equal(d, c, decimal=2)


if __name__ == '__main__':
    unt.main()




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Contains all functions required to calculate features of MC.
All of these are implemented using numba for speed.
Created on Fri Aug 16 16:59:24 2019

@author: givanna
"""

from numba import jit
import numpy as np

@jit(nopython=True)
def calculate_squared_variance(cf1, cf2, cum_weight):

    cf2_over_weight = np.divide(cf2, cum_weight)
    cf1_over_weight = np.square(np.divide(cf1, cum_weight))

    squared_variance = np.subtract(cf2_over_weight, cf1_over_weight)

    return squared_variance

@jit(nopython=True)
def update_cf(cf1, cf2, new_point):
    new_cf1 = np.add(cf1, new_point)
    new_cf2 = np.add(cf2, np.square(new_point))

    return new_cf1, new_cf2

@jit(nopython=True)
def calculate_centroid(cf1, cum_weight):
    return np.divide(cf1, cum_weight)

@jit(nopython=True)
def calculate_projected_distance(mc_centroid, mc_pref_dim, new_point):
    point_minus_centroid = np.subtract(new_point, mc_centroid)
    point_minus_centroid_sq = np.square(point_minus_centroid)
    division_result = np.divide(point_minus_centroid_sq, mc_pref_dim)

    distance = np.sum(division_result)

    return distance

@jit(nopython=True)
def calculate_projected_radius_squared(cf1, cf2, pref_dim, cum_weight):
    c2_over_weight = np.divide(cf2, cum_weight)
    c1_over_weight = np.divide(cf1, cum_weight)
    c1_over_weight_sq = np.square(c1_over_weight)

    c2_minus_c1 = np.subtract(c2_over_weight, c1_over_weight_sq)
    c2_minus_c1_over_pdim = np.divide(c2_minus_c1, pref_dim)

    radius_squared = np.sum(c2_minus_c1_over_pdim)

    return radius_squared

@jit(nopython=True)
def clone_cf(cf1, cf2):
    new_cf1 = np.copy(cf1)
    new_cf2 = np.copy(cf2)
    return new_cf1, new_cf2

@jit(nopython=True)
def is_core(cf1, cf2, pref_dim, cum_weight,
            radius_threshold_squared, density_threshold,
            max_subspace_dimensionality):
    radius_squared = calculate_projected_radius_squared(cf1, cf2, pref_dim, cum_weight)

    # the output is a tuple. We only want the count. Hence the weird format.
    # see: https://docs.scipy.org/doc/numpy-1.13.0/reference/generated/numpy.where.html
    cnt_pdim = len(np.where(pref_dim > 1)[0])

    core_status = (radius_squared <= radius_threshold_squared
            and cum_weight >= density_threshold
            and cnt_pdim <= max_subspace_dimensionality)
    return core_status



import copy as cp


# -*- coding: utf-8 -*-
class Cluster(object):
    def __init__(self, pcore_ids, cluster_centroid=None, cumulative_weight=None, preferred_dimensions=None):
        """
        Cluster object mainly used in tracking

        Args:
            pcore_ids (list): list of the id of pcore microclusters to be included in the cluster.
            cluster_centroid (list): list containing cluster centroids.
            cumulative_weight (Decimal): weight of the cluster.
            preferred_dimensions (numpy.array): cluster's preferred dimensions.
        """
        self.pcore_ids = pcore_ids
        self.id = set()
        # Be wary of this changing to string after children is assigned as parents.
        self.parents = set()

        # These are only used for printing the result.
        self.centroid = cluster_centroid
        self.cumulative_weight = cumulative_weight

        # Used by track by historical association
        self.pcore_objects = []
        self.historical_associates = set()

        # For find the closest gate
        self.preferred_dimensions = preferred_dimensions

        # This is only to find out which pcore is the closest when adding historical associate
        self.historical_associates_pcores = set()

    def add_id(self, id):
        self.id.update([id])

    def add_parent(self, id):
        self.parents.update([id])

    def set_parents(self, parent_pcores_to_id):
        pcore_ids = self.pcore_ids
        for pcore in pcore_ids:
            if pcore in parent_pcores_to_id:
                self.add_parent(parent_pcores_to_id[pcore])

    def get_parents(self):
        return self.parents

    def add_pcore_objects(self, pcore_id_to_object):
        pcore_ids = self.pcore_ids

        for pcore_id in pcore_ids:
            pcore = pcore_id_to_object[pcore_id]

            # try using deep copy
            pcore_copy = cp.deepcopy(pcore)

            # pcore_copy = pcore.get_copy()
            # pcore_copy.preferred_dimension_vector = np.zeros(
            #     len(pcore.preferred_dimension_vector)) + pcore.preferred_dimension_vector
            # pcore_copy.set_centroid()
            # pcore_copy.id = pcore.id
            # pcore_copy.points = pcore.points[:]
            self.pcore_objects.append(pcore_copy)

    def add_historical_associate(self, associate):
        self.historical_associates.update([associate])

    def add_historical_associate_pcore(self, pcore_id):
        """
        This is only to find out which pcore is actually the closest when adding a historical associate.

        :param pcore_id: the pcore id
        """
        self.historical_associates_pcores.update(pcore_id)

    def get_historical_associates_as_str(self):
        hist_assoc = sorted(self.historical_associates)
        return '&'.join(str(s) for s in hist_assoc)

    def get_historical_associates_pcore_as_str(self):
        hist_assoc_pcores = self.historical_associates_pcores
        return '&'.join(str(s) for s in hist_assoc_pcores)

    def get_preferred_dimensions_as_str(self):
        pref_dims = ';'.join(str(s) for s in self.preferred_dimensions)
        return pref_dims

    def get_pcore_ids_as_str(self):
        pcore_ids = '|'.join(str(s) for s in self.pcore_ids)
        return pcore_ids

    def get_projected_dist_to_point(self, other_point):
        """
        This is exact copy of the microcluster version.
        TODO Merge this with microcluster.
        """
        centroid = self.centroid
        pref_dim = self.preferred_dimensions
        dist = 0.0
        for c_i, p_i, d_i in zip(centroid, other_point, pref_dim):
            dist += ((float(p_i) - float(c_i)) ** 2) / float(d_i)
        return dist

    def get_dist_to_point(self, other_point):
        """
        This is exact copy of the microcluster version.
        TODO Merge this with microcluster.
        """
        centroid = self.centroid
        dist = 0.0
        for i, cluster_centroid in enumerate(centroid):
            dist += ((float(other_point[i]) - float(cluster_centroid)) ** 2)
        return dist



import numpy as np
from chronoclust.objects.microcluster import Microcluster


class PredeconMC(Microcluster):

    def __init__(self, centroid, id, is_core_cluster, cluster_CF1, cluster_CF2,
                 cluster_cumulative_weight):
        """
        This class implement both MicroCluster and Datapoint. It is being used in offline clustering step where PreDeCon is run on every pcore cluster, treating each pcore cluster as a datapoint.

        Args:
            centroid (numpy.array): datapoint's value.
            datapoint_id (:obj:set): id of the datapoint.
            is_core_cluster (bool): is this a core.
            cluster_CF1 (numpy.array): CF1 value. See Microcluster class.
            cluster_CF2 (numpy.array): CF2 value. See Microcluster class.
            cluster_cummulative_weight (float): Cummulative weight of the cluster.
        """

        self.centroid = centroid

        # List of points' id that are within certain distance from this point. Distance is measured using
        # preferred weighted similarity. See definition 3, 4, 5 in paper [2].
        self.weighted_neighbour_pts = []

        # List containing all the datapoints who are this datapoint's neighbours
        self.neighbour_pts = []

        # List representing subspace preference of a point. See definition 3 in paper [2]. A dimension is 1
        # in the vector if it is preferred by the point i.e. the variance in the dimension is smaller than the
        # variance threshold (delta in paper [2)], k otherwise where k is just a constant.
        self.subspace_preference_vector = np.ones(len(centroid))

        # A status assigned to this data point.
        # 'c' means classified, 'u' means unclassified, 'n' means noise.
        self._classification = 'u'
        self.core_status = is_core_cluster
        self.id = id
        Microcluster.__init__(self, cf1=cluster_CF1, cf2=cluster_CF2,
                              cumulative_weight=cluster_cumulative_weight,
                              id=id)

    def get_centroid(self):
        return self.centroid

    def is_core(self):
        return self.core_status

    def merge_mc(self, other_mc):
        """
        Add this cluster to another cluster (supposedly core cluster).

        Args:
            other_mc (Microcluster): Cluster to add this cluster into.

        Returns:
            None.
        """
        # me_as_microcluster = Microcluster(cf1=np.copy(self.CF1), cf2=np.copy(self.CF2), id=self.id,
        #                                   cumulative_weight=self.cumulative_weight)
        # other_mc.add_new_cluster(me_as_microcluster)

        other_mc.CF1 += self.CF1
        other_mc.CF2 += self.CF2
        other_mc.cumulative_weight += self.cumulative_weight
        other_mc.id.add(self.id)
        other_mc.set_centroid()

    def get_pdim(self):
        """
        Calculate the preference dimensionality (PDim in paper[2]) i.e. number of dimensions with variance less than
        delta. See definition 2 in paper[2].
        Method does so by converting subspace_preference_vector into array of boolean i.e. the one with >1 value is
        true, otherwise false, then sum the array up (assuming true = 1, false = 0).

        Returns:
            Int: Preference dimensionality.
        """
        return (np.array(self.subspace_preference_vector) > 1).sum()

    def is_classified(self):
        return self._classification == 'c'

    def is_noise(self):
        return self._classification == 'n'

    def is_unclassified(self):
        return self._classification == 'u'

    def set_classified(self):
        self._classification = 'c'

    def set_noise(self):
        self._classification = 'n'



#!/usr/bin/env python
"""
Contain all the objects used in both hddstream and PreDeCon algorithm.
"""

import numpy as np
import chronoclust.utilities.mc_functions as nmba

__author__ = "Givanna Putri, Deeksha Singh, Mark Read, and Tao Tang"
__copyright__ = "Copyright 2017, Cytoclust Project"
__credits__ = ["Givanna Putri", "Deeksha Singh", "Mark Read", "Tao Tang"]
__version__ = "0.0.1"
__maintainer__ = "Givanna Putri"
__email__ = "ghar1821@uni.sydney.edu.au"
__status__ = "Development"


class Microcluster(object):
    def __init__(self, cf1, cf2, id=set(), cumulative_weight=0,
                 preferred_dimension_vector=None, cluster_centroids=None, creation_time_in_hrs=0):
        """
        Class representing Microcluster (MC) in Chronoclust

        Parameters
        ----------
        cf1 : np.ndarray
            (Cluster Feature 1) weighted linear sum of all points in this Microcluster for each dimension.
        cf2 : np.ndarray
            (Cluster Feature 2) weighted linear sum of square of all points in this Microcluster for each dimension.
        id : np.ndarray, optional
            id of the Microcluster. Array containing single int for potential and outlier,
            multiple ints for core(consists of the ids of all constituting potential Microcluster).
        cumulative_weight : float, optional
            Sum of the weight of the datapoints in the cluster over time.
            For our usage, each datapoint worth weight of 1.
        preferred_dimension_vector : np.ndarray, optional
            An array which each index indicates whether the dimension is preferred by the Microcluster.
        cluster_centroids : np.ndarray, optional
            Cluster centroid.
        creation_time_in_hrs : int, optional
            Time when the cluster is created in hours.

        Attributes
        ----------
        CF1 : np.ndarray
            (Cluster Feature 1) weighted linear sum of all points in this Microcluster for each dimension.
            Don't get confused with the paper definition of f(t) * pij where f(t) is weight of the datapoint at
            time t when adding new data point. The way we do it is fine because rather than compunding the decay
            rate based on the arrival time, we just apply it again to the new weight at every time interval.
            So for instance, CF1 at t0 is 7, at t1 it's decayed to 3.5 by multiplying 7 with 2^-1 assuming lambda is 1.
            Then at t2, rather than multiplying 7 with 2^-2, we multiply 3.5 by 2^-1 again, yielding same value 1.75.
        CF2 : np.ndarray
            (Cluster Feature 2) weighted linear sum of square of all points in this Microcluster for each dimension.
        id : np.ndarray
            id of the Microcluster. Array containing single int for potential and outlier,
            multiple ints for core(consists of the ids of all constituting potential Microcluster).
        cumulative_weight : float
            Sum of the weight of the datapoints in the cluster over time.
            For our usage, each datapoint worth weight of 1.
        preferred_dimension_vector : np.ndarray
            An array which each index indicates whether the dimension is preferred by the Microcluster.
        cluster_centroids : np.ndarray
            Cluster centroid.
        creation_time_in_hrs : int
            Time when the cluster is created in hours.
        points : dict
            All the datapoints within the Microcluster. Key is the points id (assigned when adding new points).
            Value is the numerical value of the data point.
        """

        self.id = id
        self.CF1 = cf1
        self.CF2 = cf2
        self.cumulative_weight = cumulative_weight
        self.preferred_dimension_vector = preferred_dimension_vector
        self.cluster_centroids = cluster_centroids
        self.creation_time_in_hrs = creation_time_in_hrs
        self.points = {}

        self.prev_pcore_id = None
        self.prev_outlier_id = None

    def update_prev_outlier_id(self, outlier_id):
        self.prev_outlier_id = outlier_id

    def update_prev_pcore_id(self, pcore_id):
        self.prev_pcore_id = pcore_id

    def update_preferred_dimensions(self, variance_threshold_squared, k_constant):
        """
        Calculate the preferred dimensions of the cluster. When calculating the preferred dimensions, the method
        compares squared variance along each dimension with squared variance threshold.

        Args:
            variance_threshold_squared (float): The squared variance threshold used to determine whether a dimension
                can be considered preferred. For a dimension to be preferred, the variance of the data points along that
                dimension must be less than or equal to the the threshold.
            k_constant (int): Default constant assigned to a dimension whose variance is smaller than variance_threshold.

        Returns:
            None.
        """

        cf1 = self.CF1
        cf2 = self.CF2
        cum_weight = self.cumulative_weight

        squared_variance = nmba.calculate_squared_variance(cf1, cf2, cum_weight)
        updated_pref_dim_vector = []
        for s in squared_variance:
            if s <= variance_threshold_squared:
                updated_pref_dim_vector.append(k_constant)
            else:
                updated_pref_dim_vector.append(1.0)
        self.preferred_dimension_vector = np.array(updated_pref_dim_vector)

    def add_new_point(self, new_point_values, new_point_timestamp, new_point_idx, new_point_weight=1,
                      update_centroid=True):

        """
        Add new point to the microcluster. In our usage, each point is initially of weight 1. This makes sum of
        weight to be the same as number of points.

        Parameters
        ----------
        new_point_values : np.ndarray
            The datapoint represented as an array of value.
        new_point_timestamp : int
            The timepoint the datapoint is meant for.
        new_point_idx : int
            The index (or so called id) of the data point.
        new_point_weight : int, optional
            Weight of the datapoint to be added. Default to 1.
        update_centroid : bool, optional
            Whether to update the MC's centroid after adding new point. Default to True.
            This parameter is mainly used for testing for now.

        Returns
        -------
        None

        """

        cf1 = self.CF1
        cf2 = self.CF2
        self.CF1, self.CF2 = nmba.update_cf(cf1, cf2, new_point_values)
        self.cumulative_weight += new_point_weight

        self.points[new_point_idx] = new_point_values.tolist()

        # update the cluster centroid as it may have moved with the introduction of new data point.
        if update_centroid:
            self.set_centroid()

    def set_centroid(self):
        """
        Calculate and set the microcluster's centroid.

        Returns
        -------
        None
        """
        cf1 = self.CF1
        cum_weight = self.cumulative_weight
        self.cluster_centroids = nmba.calculate_centroid(cf1, cum_weight)

    def get_projected_dist_to_point(self, other_point):
        """
        Calculate the projected distance between this cluster and a datapoint p. See Definition 8 of paper[1]

        Args:
            other_point (numpy.ndarray): The datapoint represented as an array of value of each of its dimension.

        Returns:
            Float: Projected distance between this point and point given as argument.
        """
        centroid = np.array(self.cluster_centroids)
        pref_dim = np.array(self.preferred_dimension_vector)
        other_pt_np = np.array(other_point)
        distance = nmba.calculate_projected_distance(centroid, pref_dim, other_pt_np)
        return distance

    def calculate_projected_radius_squared(self):
        """
        Calculate projected radius. See definition 5 in paper[1]. Note that this calculate SQUARED radius.

        Returns:
            Float: Squared Projected radius.
        """
        cf1 = self.CF1
        cf2 = self.CF2
        pref_dim = self.preferred_dimension_vector
        cum_weight = self.cumulative_weight

        radius_squared = nmba.calculate_projected_radius_squared(cf1, cf2, pref_dim, cum_weight)

        return radius_squared

    def get_copy(self):
        """
        Return a clone of itself. Cannot use copy or deepcopy as they don't work for class object.

        Returns:
            Microcluster: A clone of itself.
        """

        cf1 = self.CF1
        cf2 = self.CF2

        new_cf1, new_cf2 = nmba.clone_cf(cf1, cf2)
        return Microcluster(cf1=new_cf1, cf2=new_cf2, cumulative_weight=self.cumulative_weight)

    def get_copy_with_new_point(self, datapoint, variance_threshold_squared, k_constant):
        """
        Return a clone of itself with a datapoint added in it. It will create a clone of itself (note it'll be a
        standalone clone as CF1 and CF2 will not be copied over. Beware of Python assignment is passing pointers!),
        then add a new datapoint to it, then finally recalculate the preferred_dimension_vector, and return it.
        This is used to simulate adding a point to a cluster to see if the cluster can fit another point.

        Args:
            datapoint (numpy.array): A datapoint represented as array of values in each dimension, to be added to the
                clone of this cluster.
            variance_threshold_squared (float): Variance_threshold used to calculate preferred_dimension_vector.
            k_constant (int): k_constant used to calculate preferred_dimension_vector.

        Returns:
            Microcluster: A clone of itself with new datapoint added in it.
        """
        temp_pmc = self.get_copy()
        temp_pmc.add_new_point(datapoint, -1, -1)
        temp_pmc.update_preferred_dimensions(variance_threshold_squared, k_constant)

        return temp_pmc

    def is_core(self, radius_threshold_squared, density_threshold, max_subspace_dimensionality):
        """
        Check if this cluster is a core cluster. See definition 4 in paper [1].

        Args:
            radius_threshold_squared (float): Squared minimum projected radius.
            density_threshold (float): Squared maximum density threshold.
            max_subspace_dimensionality (int): Minimum number of dimensions in pdim with value k_constant.

        Returns:
            bool: True if cluster is a core. False otherwise.
        """
        # TODO: improve the core radius calculation based on decimal point?
        cf1 = self.CF1
        cf2 = self.CF2
        cum_weight = self.cumulative_weight
        pref_dim = self.preferred_dimension_vector

        core_status = nmba.is_core(cf1, cf2, pref_dim, cum_weight,
                                   radius_threshold_squared, density_threshold,
                                   max_subspace_dimensionality)
        return core_status

    def reset_points(self):
        # TODO revisit this
        self.points = {}



import unittest as unt

from chronoclust.objects.cluster import Cluster
import chronoclust.tracking.cluster_tracker as ct


class TrackByLineageTest(unt.TestCase):

    def test_new_letter_assignment(self):
        tracker = ct.TrackByLineage()
        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1']))
        tracker.add_new_child_cluster(Cluster(['2', '3']))

        tracker.calculate_ids()
        self.assertEqual(['A', 'B', 'C'], [t.id for t in tracker.child_clusters])

    def test_new_letter_assignment_beyond_26_clusters(self):
        tracker = ct.TrackByLineage()
        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1']))
        tracker.add_new_child_cluster(Cluster(['2']))
        tracker.add_new_child_cluster(Cluster(['3']))
        tracker.add_new_child_cluster(Cluster(['4']))
        tracker.add_new_child_cluster(Cluster(['5']))
        tracker.add_new_child_cluster(Cluster(['6']))
        tracker.add_new_child_cluster(Cluster(['7']))
        tracker.add_new_child_cluster(Cluster(['8']))
        tracker.add_new_child_cluster(Cluster(['9']))
        tracker.add_new_child_cluster(Cluster(['10']))
        tracker.add_new_child_cluster(Cluster(['11']))
        tracker.add_new_child_cluster(Cluster(['12']))
        tracker.add_new_child_cluster(Cluster(['13']))
        tracker.add_new_child_cluster(Cluster(['14']))
        tracker.add_new_child_cluster(Cluster(['15']))
        tracker.add_new_child_cluster(Cluster(['16']))
        tracker.add_new_child_cluster(Cluster(['17']))
        tracker.add_new_child_cluster(Cluster(['18']))
        tracker.add_new_child_cluster(Cluster(['19']))
        tracker.add_new_child_cluster(Cluster(['20']))
        tracker.add_new_child_cluster(Cluster(['21']))
        tracker.add_new_child_cluster(Cluster(['22']))
        tracker.add_new_child_cluster(Cluster(['23']))
        tracker.add_new_child_cluster(Cluster(['24']))
        tracker.add_new_child_cluster(Cluster(['25']))
        tracker.add_new_child_cluster(Cluster(['26']))

        tracker.calculate_ids()
        self.assertEqual(['A', 'B', 'C', 'D', 'E', 'F', 'G', 'H', 'I',
                          'J', 'K', 'L', 'M', 'N', 'O', 'P', 'Q', 'R',
                          'S', 'T', 'U', 'V', 'W', 'X', 'Y', 'Z', 'AA'],
                         [t.id for t in tracker.child_clusters])

    def test_new_clusters_at_end(self):
        tracker = ct.TrackByLineage()
        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1']))
        tracker.add_new_child_cluster(Cluster(['2', '3']))

        tracker.calculate_ids()
        tracker_result = [t.id for t in tracker.child_clusters]
        tracker.transfer_child_to_parent()

        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1']))
        tracker.add_new_child_cluster(Cluster(['2', '3']))
        tracker.add_new_child_cluster(Cluster(['4']))
        tracker.add_new_child_cluster(Cluster(['5', '6']))

        tracker.calculate_ids()
        tracker_result.extend([t.id for t in tracker.child_clusters])

        self.assertEqual(['A', 'B', 'C', 'A', 'B', 'C', 'D', 'E'], tracker_result)

    def test_new_clusters_at_middle(self):
        tracker = ct.TrackByLineage()
        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1']))
        tracker.add_new_child_cluster(Cluster(['2', '3']))

        tracker.calculate_ids()
        tracker_result = [t.id for t in tracker.child_clusters]
        tracker.transfer_child_to_parent()

        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1']))
        tracker.add_new_child_cluster(Cluster(['4', '5']))
        tracker.add_new_child_cluster(Cluster(['2', '3']))

        tracker.calculate_ids()
        tracker_result.extend([t.id for t in tracker.child_clusters])

        self.assertEqual(['A', 'B', 'C', 'A', 'B', 'D', 'C'], tracker_result)

    def test_simple_split(self):
        tracker = ct.TrackByLineage()
        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1', '2', '3']))

        tracker.calculate_ids()
        tracker_result = [t.id for t in tracker.child_clusters]
        tracker.transfer_child_to_parent()

        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1', '2']))
        tracker.add_new_child_cluster(Cluster(['3']))

        tracker.calculate_ids()
        tracker_result.extend([t.id for t in tracker.child_clusters])

        self.assertEqual(['A', 'B', 'A', 'B', 'B|1'], tracker_result)

    def test_simple_split_and_new(self):
        tracker = ct.TrackByLineage()
        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1', '2', '3']))

        tracker.calculate_ids()
        tracker_result = [t.id for t in tracker.child_clusters]
        tracker.transfer_child_to_parent()

        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1', '2']))
        tracker.add_new_child_cluster(Cluster(['3']))
        tracker.add_new_child_cluster(Cluster(['4']))

        tracker.calculate_ids()
        tracker_result.extend([t.id for t in tracker.child_clusters])

        self.assertEqual(['A', 'B', 'A', 'B', 'B|1', 'C'], tracker_result)

    def test_majority_later_split_assignment(self):
        tracker = ct.TrackByLineage()
        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1', '2', '3']))

        tracker.calculate_ids()
        tracker_result = [t.id for t in tracker.child_clusters]
        tracker.transfer_child_to_parent()

        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1']))
        tracker.add_new_child_cluster(Cluster(['2', '3']))

        tracker.calculate_ids()
        tracker_result.extend([t.id for t in tracker.child_clusters])

        self.assertEqual(['A', 'B', 'A', 'B|1', 'B'], tracker_result)

    def test_majority_first_split_assignment(self):
        tracker = ct.TrackByLineage()
        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1', '2', '3']))

        tracker.calculate_ids()
        tracker_result = [t.id for t in tracker.child_clusters]
        tracker.transfer_child_to_parent()

        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1', '2']))
        tracker.add_new_child_cluster(Cluster(['3']))

        tracker.calculate_ids()
        tracker_result.extend([t.id for t in tracker.child_clusters])

        self.assertEqual(['A', 'B', 'A', 'B', 'B|1'], tracker_result)

    def test_majority_middle_split_assignment(self):
        tracker = ct.TrackByLineage()
        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1', '2', '3', '4']))

        tracker.calculate_ids()
        tracker_result = [t.id for t in tracker.child_clusters]
        tracker.transfer_child_to_parent()

        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1']))
        tracker.add_new_child_cluster(Cluster(['2', '3']))
        tracker.add_new_child_cluster(Cluster(['4']))

        tracker.calculate_ids()
        tracker_result.extend([t.id for t in tracker.child_clusters])

        self.assertEqual(['A', 'B', 'A', 'B|1', 'B', 'B|2'], tracker_result)

    def test_multi_day_split_assignment(self):
        tracker = ct.TrackByLineage()
        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1', '2', '3']))

        tracker.calculate_ids()
        tracker_result = [t.id for t in tracker.child_clusters]
        tracker.transfer_child_to_parent()

        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1', '2']))
        tracker.add_new_child_cluster(Cluster(['3']))

        tracker.calculate_ids()
        tracker_result.extend([t.id for t in tracker.child_clusters])
        tracker.transfer_child_to_parent()

        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1']))
        tracker.add_new_child_cluster(Cluster(['2']))
        tracker.add_new_child_cluster(Cluster(['3']))
        tracker.calculate_ids()
        tracker_result.extend([t.id for t in tracker.child_clusters])

        self.assertEqual(['A', 'B', 'A', 'B', 'B|1', 'A', 'B', 'B|2', 'B|1'], tracker_result)

    def test_complicated_split(self):
        tracker = ct.TrackByLineage()
        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1']))
        tracker.add_new_child_cluster(Cluster(['2', '3']))
        tracker.add_new_child_cluster(Cluster(['4', '5', '6', '7', '8', '9']))

        tracker.calculate_ids()
        tracker_result = [t.id for t in tracker.child_clusters]
        tracker.transfer_child_to_parent()

        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1']))
        tracker.add_new_child_cluster(Cluster(['2', '3']))
        tracker.add_new_child_cluster(Cluster(['4', '5', '6']))
        tracker.add_new_child_cluster(Cluster(['7', '8', '9']))

        tracker.calculate_ids()
        tracker_result.extend([t.id for t in tracker.child_clusters])
        tracker.transfer_child_to_parent()

        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1']))
        tracker.add_new_child_cluster(Cluster(['2', '3']))
        tracker.add_new_child_cluster(Cluster(['4', '5', '6']))
        tracker.add_new_child_cluster(Cluster(['7']))
        tracker.add_new_child_cluster(Cluster(['8', '9']))

        tracker.calculate_ids()
        tracker_result.extend([t.id for t in tracker.child_clusters])
        tracker.transfer_child_to_parent()

        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1']))
        tracker.add_new_child_cluster(Cluster(['2', '3'])),
        tracker.add_new_child_cluster(Cluster(['4', '5', '6']))
        tracker.add_new_child_cluster(Cluster(['7']))
        tracker.add_new_child_cluster(Cluster(['8']))
        tracker.add_new_child_cluster(Cluster(['9']))

        tracker.calculate_ids()
        tracker_result.extend([t.id for t in tracker.child_clusters])
        tracker.transfer_child_to_parent()

        self.assertEqual(['A', 'B', 'C', 'D',
          'A', 'B', 'C', 'D', 'D|1',
          'A', 'B', 'C', 'D', 'D|1|1', 'D|1',
          'A', 'B', 'C', 'D', 'D|1|1', 'D|1', 'D|1|2'], tracker_result)

    def test_merge(self):
        tracker = ct.TrackByLineage()

        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1']))
        tracker.add_new_child_cluster(Cluster(['2', '3']))
        tracker.add_new_child_cluster(Cluster(['4']))

        tracker.calculate_ids()
        tracker_result = [t.id for t in tracker.child_clusters]
        tracker.transfer_child_to_parent()

        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1', '2', '3']))
        tracker.add_new_child_cluster(Cluster(['4']))

        tracker.calculate_ids()
        tracker_result.extend([t.id for t in tracker.child_clusters])
        tracker.transfer_child_to_parent()

        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1', '2', '3', '4']))

        tracker.calculate_ids()
        tracker_result.extend([t.id for t in tracker.child_clusters])
        tracker.transfer_child_to_parent()

        self.assertEqual(['A', 'B', 'C', 'D', 'A', '(B,C)', 'D', 'A', '((B,C),D)'], tracker_result)

    def test_merge_then_split(self):
        tracker = ct.TrackByLineage()

        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1']))
        tracker.add_new_child_cluster(Cluster(['2', '3']))

        tracker.calculate_ids()
        tracker_result = [t.id for t in tracker.child_clusters]
        tracker.transfer_child_to_parent()

        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1', '2', '3']))

        tracker.calculate_ids()
        tracker_result.extend([t.id for t in tracker.child_clusters])
        tracker.transfer_child_to_parent()

        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1', '2']))
        tracker.add_new_child_cluster(Cluster(['3']))
        tracker.add_new_child_cluster(Cluster(['4']))

        tracker.calculate_ids()
        tracker_result.extend([t.id for t in tracker.child_clusters])
        tracker.transfer_child_to_parent()

        self.assertEqual(['A', 'B', 'C', 'A', '(B,C)', 'A', '(B,C)', '(B,C)|1', 'D'], tracker_result)

    def test_merge_and_split(self):
        tracker = ct.TrackByLineage()

        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1']))
        tracker.add_new_child_cluster(Cluster(['2', '3']))

        tracker.calculate_ids()
        tracker_result = [t.id for t in tracker.child_clusters]
        tracker.transfer_child_to_parent()

        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1', '2', '3']))
        tracker.add_new_child_cluster(Cluster(['4']))

        tracker.calculate_ids()
        tracker_result.extend([t.id for t in tracker.child_clusters])
        tracker.transfer_child_to_parent()

        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1', '2']))
        tracker.add_new_child_cluster(Cluster(['3', '4']))

        tracker.calculate_ids()
        tracker_result.extend([t.id for t in tracker.child_clusters])
        tracker.transfer_child_to_parent()

        self.assertEqual(['A', 'B', 'C', 'A', '(B,C)', 'D', 'A', '(B,C)', '((B,C)|1,D)'], tracker_result)

    def test_split_equally(self):
        tracker = ct.TrackByLineage()

        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1']))

        tracker.calculate_ids()
        tracker_result = [t.id for t in tracker.child_clusters]
        tracker.transfer_child_to_parent()

        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1', '2', '3']))

        tracker.calculate_ids()
        tracker_result.extend([t.id for t in tracker.child_clusters])
        tracker.transfer_child_to_parent()

        tracker.add_new_child_cluster(Cluster(['0']))
        tracker.add_new_child_cluster(Cluster(['1']))
        tracker.add_new_child_cluster(Cluster(['2']))
        tracker.add_new_child_cluster(Cluster(['3']))

        tracker.calculate_ids()
        tracker_result.extend([t.id for t in tracker.child_clusters])
        tracker.transfer_child_to_parent()

        self.assertEqual(['A', 'B', 'A', 'B', 'A', 'B', 'B|1', 'B|2'], tracker_result)

    def test_split_with_merge_and_new(self):
        tracker = ct.TrackByLineage()

        tracker.add_new_child_cluster(Cluster(['1', '2', '4']))
        tracker.calculate_ids()
        tracker_result = [t.id for t in tracker.child_clusters]
        tracker.transfer_child_to_parent()

        tracker.add_new_child_cluster(Cluster(['1', '2', '4', '6', '7', '9', '15']))
        tracker.calculate_ids()
        tracker_result.extend([t.id for t in tracker.child_clusters])
        tracker.transfer_child_to_parent()

        tracker.add_new_child_cluster(Cluster(['1', '2', '4', '9', '15', '24', '25']))
        tracker.add_new_child_cluster(Cluster(['26', '27', '21', '6']))
        tracker.add_new_child_cluster(Cluster(['23', '7']))

        tracker.calculate_ids()
        tracker_result.extend([t.id for t in tracker.child_clusters])
        tracker.transfer_child_to_parent()

        tracker.add_new_child_cluster(Cluster(['1', '2', '33', '4', '6', '39', '9', '15', '21', '24', '25']))
        tracker.add_new_child_cluster(Cluster(['34', '38', '7', '40', '23']))
        tracker.add_new_child_cluster(Cluster(['26', '27']))

        tracker.calculate_ids()
        tracker_result.extend([t.id for t in tracker.child_clusters])
        tracker.transfer_child_to_parent()

        self.assertEqual(['A', 'A', 'A', 'A|1', 'A|2', '(A,A|1)', 'A|2', 'A|1|1'], tracker_result)


if __name__ == '__main__':
    unt.main()



import unittest as unt
import numpy as np

from chronoclust.objects.cluster import Cluster
from chronoclust.objects.microcluster import Microcluster
import chronoclust.tracking.cluster_tracker as ct


class TrackByHistAssocTest(unt.TestCase):

    def test_no_previous_day_cluster(self):
        """
        Test what happen when there is no cluster found on previous day.
        Expected outcome: all clusters historical associate set to None
        """
        tracker = ct.TrackByHistoricalAssociation()
        cluster = Cluster([0, 1, 2])
        pcore0 = Microcluster(cf1=np.array([1, 1]), cf2=np.array([1, 1]), id=0,
                              preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([1, 1]))
        pcore1 = Microcluster(cf1=np.array([2, 2]), cf2=np.array([4, 4]), id=1,
                              preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([2, 2]))
        pcore2 = Microcluster(cf1=np.array([3, 3]), cf2=np.array([9, 9]), id=2,
                              preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([3, 3]))
        pcore_id_to_obj = {
            0: pcore0, 1: pcore1, 2: pcore2
        }
        cluster.add_pcore_objects(pcore_id_to_obj)
        tracker.current_clusters.append(cluster)
        tracker.track_cluster_history()

        self.assertEqual(1, len(cluster.historical_associates))
        self.assertIsNone(cluster.historical_associates.pop())

    def test_pcore_moves_a_bit(self):
        """
        Test what happen when the cluster on next day have pcores shifted slightly
        Expected outcome: all clusters historical associate have same id as clusters in previous day
        """
        tracker_by_association = ct.TrackByHistoricalAssociation()
        tracker_lineage = ct.TrackByLineage()

        cluster1 = Cluster([0, 1])
        cluster2 = Cluster([2, 3])
        pcore0 = Microcluster(cf1=np.array([0.1, 0.1]), cf2=np.array([0.01, 0.01]), id={0},
                              preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.1, 0.1]))
        pcore1 = Microcluster(cf1=np.array([0.2, 0.2]), cf2=np.array([0.04, 0.04]), id={1},
                              preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.2, 0.2]))
        pcore2 = Microcluster(cf1=np.array([0.8, 0.8]), cf2=np.array([0.64, 0.64]), id={2},
                              preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.8, 0.8]))
        pcore3 = Microcluster(cf1=np.array([0.9, 0.9]), cf2=np.array([0.81, 0.81]), id={3},
                              preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.9, 0.9]))
        cluster1.add_pcore_objects({0: pcore0, 1: pcore1})
        cluster2.add_pcore_objects({2: pcore2, 3: pcore3})

        tracker_lineage.add_new_child_cluster(cluster1)
        tracker_lineage.add_new_child_cluster(cluster2)
        tracker_lineage.calculate_ids()

        tracker_by_association.set_current_clusters(tracker_lineage.child_clusters)
        tracker_by_association.track_cluster_history()

        # Prepare for the next time point
        tracker_lineage.transfer_child_to_parent()
        tracker_by_association.transfer_current_to_previous()

        # next time point
        clusters = [Cluster([0, 1]), Cluster([2, 3])]
        pcores = [
            Microcluster(cf1=np.array([0.11, 0.11]), cf2=np.array([0.021, 0.021]), id={0},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.11, 0.11])),
            Microcluster(cf1=np.array([0.21, 0.21]), cf2=np.array([0.0441, 0.0441]), id={1},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.21, 0.21])),
            Microcluster(cf1=np.array([0.81, 0.81]), cf2=np.array([0.6561, 0.6561]), id={2},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.81, 0.81])),
            Microcluster(cf1=np.array([0.91, 0.91]), cf2=np.array([0.8281, 0.8281]), id={3},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.91, 0.91]))
        ]

        clusters[0].add_pcore_objects({0: pcores[0], 1: pcores[1]})
        clusters[1].add_pcore_objects({2: pcores[2], 3: pcores[3]})
        for cl in clusters:
            tracker_lineage.add_new_child_cluster(cl)
        tracker_lineage.calculate_ids()

        tracker_by_association.set_current_clusters(tracker_lineage.child_clusters)
        tracker_by_association.track_cluster_history()

        self.assertEqual(1, len(clusters[0].historical_associates))
        self.assertEqual(1, len(clusters[1].historical_associates))

        self.assertEqual("A", clusters[0].get_historical_associates_as_str())
        self.assertEqual("B", clusters[1].get_historical_associates_as_str())

    def test_new_cluser_close_to_A(self):
        """
        Test what happen when there is new cluster close to A.
        Expected outcome: the new cluster's associate is A
        """
        tracker_by_association = ct.TrackByHistoricalAssociation()
        tracker_lineage = ct.TrackByLineage()

        clusters = [Cluster([0, 1]), Cluster([2, 3])]
        pcores = [
            Microcluster(cf1=np.array([0.1, 0.1]), cf2=np.array([0.01, 0.01]), id={0},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.1, 0.1])),
            Microcluster(cf1=np.array([0.2, 0.2]), cf2=np.array([0.04, 0.04]), id={1},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.2, 0.2])),
            Microcluster(cf1=np.array([0.8, 0.8]), cf2=np.array([0.64, 0.64]), id={2},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.8, 0.8])),
            Microcluster(cf1=np.array([0.9, 0.9]), cf2=np.array([0.81, 0.81]), id={3},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.9, 0.9]))
        ]
        clusters[0].add_pcore_objects({0: pcores[0], 1: pcores[1]})
        clusters[1].add_pcore_objects({2: pcores[2], 3: pcores[3]})

        for cl in clusters:
            tracker_lineage.add_new_child_cluster(cl)
        tracker_lineage.calculate_ids()

        tracker_by_association.set_current_clusters(tracker_lineage.child_clusters)
        tracker_by_association.track_cluster_history()

        # Prepare for the next time point
        tracker_lineage.transfer_child_to_parent()
        tracker_by_association.transfer_current_to_previous()

        # next time point
        clusters = [Cluster([0, 1]), Cluster([2, 3]), Cluster([4, 5])]
        pcores = [
            Microcluster(cf1=np.array([0.11, 0.11]), cf2=np.array([0.021, 0.021]), id={0},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.11, 0.11])),
            Microcluster(cf1=np.array([0.21, 0.21]), cf2=np.array([0.0441, 0.0441]), id={1},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.21, 0.21])),
            Microcluster(cf1=np.array([0.81, 0.81]), cf2=np.array([0.6561, 0.6561]), id={2},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.81, 0.81])),
            Microcluster(cf1=np.array([0.91, 0.91]), cf2=np.array([0.8281, 0.8281]), id={3},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.91, 0.91])),
            Microcluster(cf1=np.array([0.3, 0.3]), cf2=np.array([0.09, 0.09]), id={4},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.3, 0.3])),
            Microcluster(cf1=np.array([0.4, 0.4]), cf2=np.array([0.16, 0.16]), id={5},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.4, 0.4]))
        ]

        clusters[0].add_pcore_objects({0: pcores[0], 1: pcores[1]})
        clusters[1].add_pcore_objects({2: pcores[2], 3: pcores[3]})
        clusters[2].add_pcore_objects({4: pcores[4], 5: pcores[5]})

        for cl in clusters:
            tracker_lineage.add_new_child_cluster(cl)
        tracker_lineage.calculate_ids()

        tracker_by_association.set_current_clusters(tracker_lineage.child_clusters)
        tracker_by_association.track_cluster_history()

        self.assertEqual("A", clusters[0].get_historical_associates_as_str())
        self.assertEqual("B", clusters[1].get_historical_associates_as_str())
        self.assertEqual("A", clusters[2].get_historical_associates_as_str())

    def test_split_cluster(self):
        """
        Test what happen when there is split cluster.
        Expected outcome: the split cluster has same associate.
        """
        tracker_by_association = ct.TrackByHistoricalAssociation()
        tracker_lineage = ct.TrackByLineage()

        # 3 clusters: A, B
        clusters = [Cluster([0, 1]), Cluster([2, 3])]
        pcores = [
            Microcluster(cf1=np.array([0.1, 0.1]), cf2=np.array([0.01, 0.01]), id={0},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.1, 0.1])),
            Microcluster(cf1=np.array([0.2, 0.2]), cf2=np.array([0.04, 0.04]), id={1},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.2, 0.2])),
            Microcluster(cf1=np.array([0.8, 0.8]), cf2=np.array([0.64, 0.64]), id={2},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.8, 0.8])),
            Microcluster(cf1=np.array([0.9, 0.9]), cf2=np.array([0.81, 0.81]), id={3},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.9, 0.9])),
        ]
        clusters[0].add_pcore_objects({0: pcores[0], 1: pcores[1]})
        clusters[1].add_pcore_objects({2: pcores[2], 3: pcores[3]})

        for cl in clusters:
            tracker_lineage.add_new_child_cluster(cl)
        tracker_lineage.calculate_ids()

        tracker_by_association.set_current_clusters(tracker_lineage.child_clusters)
        tracker_by_association.track_cluster_history()

        # Prepare for the next time point
        tracker_lineage.transfer_child_to_parent()
        tracker_by_association.transfer_current_to_previous()

        # next time point
        clusters = [Cluster([0, 1]), Cluster([2]), Cluster([3])]
        pcores = [
            Microcluster(cf1=np.array([0.11, 0.11]), cf2=np.array([0.021, 0.021]), id={0},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.11, 0.11])),
            Microcluster(cf1=np.array([0.21, 0.21]), cf2=np.array([0.0441, 0.0441]), id={1},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.21, 0.21])),
            Microcluster(cf1=np.array([0.81, 0.81]), cf2=np.array([0.6561, 0.6561]), id={2},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.81, 0.81])),
            Microcluster(cf1=np.array([0.91, 0.91]), cf2=np.array([0.8281, 0.8281]), id={3},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.91, 0.91])),
        ]

        clusters[0].add_pcore_objects({0: pcores[0], 1: pcores[1]})
        clusters[1].add_pcore_objects({2: pcores[2]})
        clusters[2].add_pcore_objects({3: pcores[3]})

        for cl in clusters:
            tracker_lineage.add_new_child_cluster(cl)
        tracker_lineage.calculate_ids()

        tracker_by_association.set_current_clusters(tracker_lineage.child_clusters)
        tracker_by_association.track_cluster_history()

        self.assertEqual("A", clusters[0].get_historical_associates_as_str())
        self.assertEqual("B", clusters[1].get_historical_associates_as_str())
        self.assertEqual("B", clusters[2].get_historical_associates_as_str())


    def test_merged_cluster(self):
        """
        Test what happen when there is merged cluster.
        Expected outcome: the merged cluster has 2 associates.
        """
        tracker_by_association = ct.TrackByHistoricalAssociation()
        tracker_lineage = ct.TrackByLineage()

        # 3 clusters: A, B, C
        clusters = [Cluster([0, 1]), Cluster([2, 3]), Cluster([4, 5])]
        pcores = [
            Microcluster(cf1=np.array([0.1, 0.1]), cf2=np.array([0.01, 0.01]), id={0},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.1, 0.1])),
            Microcluster(cf1=np.array([0.2, 0.2]), cf2=np.array([0.04, 0.04]), id={1},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.2, 0.2])),
            Microcluster(cf1=np.array([0.3, 0.3]), cf2=np.array([0.09, 0.09]), id={2},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.3, 0.3])),
            Microcluster(cf1=np.array([0.4, 0.4]), cf2=np.array([0.16, 0.16]), id={3},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.4, 0.4])),
            Microcluster(cf1=np.array([0.8, 0.8]), cf2=np.array([0.64, 0.64]), id={4},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.8, 0.8])),
            Microcluster(cf1=np.array([0.9, 0.9]), cf2=np.array([0.81, 0.81]), id={5},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.9, 0.9]))
        ]
        clusters[0].add_pcore_objects({0: pcores[0], 1: pcores[1]})
        clusters[1].add_pcore_objects({2: pcores[2], 3: pcores[3]})
        clusters[2].add_pcore_objects({4: pcores[4], 5: pcores[5]})

        for cl in clusters:
            tracker_lineage.add_new_child_cluster(cl)
        tracker_lineage.calculate_ids()

        tracker_by_association.set_current_clusters(tracker_lineage.child_clusters)
        tracker_by_association.track_cluster_history()

        # Prepare for the next time point
        tracker_lineage.transfer_child_to_parent()
        tracker_by_association.transfer_current_to_previous()

        # next time point
        clusters = [Cluster([0, 1, 2]), Cluster([3]), Cluster([4, 5])]
        pcores = [
            Microcluster(cf1=np.array([0.11, 0.11]), cf2=np.array([0.021, 0.021]), id={0},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.11, 0.11])),
            Microcluster(cf1=np.array([0.21, 0.21]), cf2=np.array([0.0441, 0.0441]), id={1},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.21, 0.21])),
            Microcluster(cf1=np.array([0.35, 0.35]), cf2=np.array([0.1225, 0.1225]), id={2},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.35, 0.35])),
            Microcluster(cf1=np.array([0.45, 0.45]), cf2=np.array([0.2025, 0.2025]), id={3},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.45, 0.45])),
            Microcluster(cf1=np.array([0.81, 0.81]), cf2=np.array([0.6561, 0.6561]), id={4},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.81, 0.81])),
            Microcluster(cf1=np.array([0.91, 0.91]), cf2=np.array([0.8281, 0.8281]), id={5},
                         preferred_dimension_vector=np.array([15, 15]), cluster_centroids=np.array([0.91, 0.91]))
        ]

        clusters[0].add_pcore_objects({0: pcores[0], 1: pcores[1], 2: pcores[2]})
        clusters[1].add_pcore_objects({3: pcores[3]})
        clusters[2].add_pcore_objects({4: pcores[4], 5: pcores[5]})

        for cl in clusters:
            tracker_lineage.add_new_child_cluster(cl)
        tracker_lineage.calculate_ids()

        tracker_by_association.set_current_clusters(tracker_lineage.child_clusters)
        tracker_by_association.track_cluster_history()

        # TODO should be A&B. Need to add some sorting.
        self.assertEqual("A&B", clusters[0].get_historical_associates_as_str())
        self.assertEqual("B", clusters[1].get_historical_associates_as_str())
        self.assertEqual("C", clusters[2].get_historical_associates_as_str())


if __name__ == '__main__':
    unt.main()




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="Chronoclust",
    version="0.2.1.2",
    author="Givanna Putri",
    author_email="ghar1821@uni.sydney.edu.au",
    description="A clustering algorithm that will perform clustering on each of a time-series of discrete datasets, and explicitly track the evolution of clusters over time.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://https://github.com/ghar1821/Chronoclust",
    packages=setuptools.find_packages(),
    install_requires=[
        'numba',
        'numpy',
        'pandas',
        'scipy',
        'scikit-learn',
        'tqdm'
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Programming Language :: Python :: 3.7'
    ],
    python_requires='>=3.6',
)



import numpy as np
from numba import jit, prange

@jit(nopython=True)
def calculate_euclidean_dist(a, b):
    """
    Calculate Euclidean distance between two points.

    Args:
        a (numpy.array): point 1.
        b (numpy.array): point 2.

    Returns:
        Float: Euclidean Distance between 2 points.
    """
    x = np.subtract(a,b)
    return np.linalg.norm(x)

@jit(nopython=True)
def calculate_variance_along_dimension(point, neighbours):
    """
    Calculate the variance of a point based on its neighbourhood.
    The variance is only calculated for a dimension.
    So make sure the point and neighbours are values for a dimension only!
    See Definition 1 in Bohm paper.

    Args:
        point: a point's value for a dimension.
        neighbours: the point's neighbours. Only give values for a dimension!

    Returns:
        Float: Variance of neighbourhood of a point along a dimension
    """

    subtractions = np.subtract(point, neighbours)
    sqr = np.square(subtractions)
    summation = np.sum(sqr)

    num_neighbour = len(neighbours)
    variance = np.divide(summation, num_neighbour)

    return variance

@jit(nopython=True)
def calculate_weighted_dist_squared(pref_vector, p, q):
    """
    Calculate the weighted distance between 2 points (dist_p in paper[2]). See definition 3 in paper[2]

    Args:
        pref_vector (numpy array): Point 1 subspace_preference_vector
        p (numpy array): Point 1.
        q (numpy array): Point 2.

    Returns:
        Float: Weighted distance between 2 points
    """
    subtraction = np.subtract(p, q)
    sqr = np.square(subtraction)
    mult = np.multiply(pref_vector, sqr)
    weight_dist_sqr = np.sum(mult)

    return weight_dist_sqr



#!/usr/bin/env python3

"""
The HDDStream algorithm. Values for some of the threshold are read from config.ini file.
See the following paper for more information:
[1] Ntoutsi, Irene, et al. "Density-based Projected Clustering over High Dimensional Data Streams." SDM. 2012.
The paper will be referred as paper[1].
"""

__author__ = "Givanna Putri, Deeksha Singh, Mark Read, and Tao Tang"
__copyright__ = "Copyright 2017, Cytoclust Project"
__credits__ = ["Givanna Putri", "Deeksha Singh", "Mark Read", "Tao Tang"]
__version__ = "0.0.1"
__maintainer__ = "Givanna Putri"
__email__ = "ghar1821@uni.sydney.edu.au"
__status__ = "Development"

import numpy as np
import sys
import logging
import io

from tqdm import tqdm
from chronoclust.clustering.predecon import PreDeCon
from chronoclust.objects.microcluster import Microcluster
from chronoclust.objects.predecon_mc import PredeconMC


class HDDStream(object):
    def __init__(self, config, logger):
        """
        Initialise hddstream object. All the attributes here are named based on Ntoutsi paper.
        Refer to the paper for more information.

        Parameters
        ----------
        config : dict
            Dictionary contain values for HDDStream's config.
        logger : logger object
            Logger object to log HDDStream's progress.
        """
        self.config = config
        self.pi = None
        self.mu = None
        self.epsilon = float(self.config['epsilon'])
        self.epsilon_squared = self.epsilon ** 2
        self.upsilon = float(self.config['upsilon']) * self.epsilon
        self.delta = self.calculate_pref_dim_variance_threshold()
        self.delta_squared = self.delta ** 2
        self.beta = float(self.config['beta'])
        self.k = float(self.config['k'])
        self.lambbda = float(self.config['lambda'])
        self.omicron = None

        # The following attributes are used in the algorithm implementation.
        self.pcore_MC = []
        self.outlier_MC = []
        self.final_clusters = []
        self.last_data_timestamp = 0
        self.dataset_dimensionality = 0
        self.logger = logger
        self.dataset_size = 0
        self.pcore_MC_last_id = 0
        self.outlier_MC_last_id = 0

        # used for logging
        self.logger = logger

    def __getstate__(self):
        """Return state values to be pickled."""
        return (self.pi, self.mu, self.epsilon, self.epsilon_squared, self.upsilon, self.delta, self.delta_squared,
                self.beta, self.k, self.lambbda, self.omicron, self.pcore_MC, self.outlier_MC,
                self.last_data_timestamp, self.dataset_dimensionality, self.dataset_size)

    def __setstate__(self, state):
        """Restore state from the unpickled state values."""
        self.pi, self.mu, self.epsilon, self.epsilon_squared, self.upsilon, self.delta, self.delta_squared, \
        self.beta, self.k, self.lambbda, self.omicron, self.pcore_MC, self.outlier_MC, self.last_data_timestamp, \
        self.dataset_dimensionality, self.dataset_size = state

        self.final_clusters = []

    def set_logger(self, logger):
        self.logger = logger

    def set_config(self, config):
        self.config = config

    def _set_dataset_dependent_parameters(self, input_dataset):
        """
        Set all the parameters whose values are dependent on the input_dataset. This include dataset_dimensionality,
        mu, delta, delta_squared, and pi.

        Args:
            input_dataset (numpy.array): 2d array containing input dataset for a given point in time.

        Returns:
            None
        """

        dataset_dim = input_dataset.shape[1]
        config = self.config

        self.dataset_dimensionality = dataset_dim

        # Set projected_dimensionality_threshold. This will be set only once.
        config_pi = float(config['pi'])
        if config_pi <= 0:
            # if config for projected_dimensionality_threshold is set to absurd size i.e. 0 or negative,
            # the dimensionality of the dataset is going to be used instead.
            self.pi = dataset_dim
        else:
            # Doesn't make sense for this to not be whole number
            self.pi = round(config_pi)

        # Setting outlier deletion point. It's given as proportion of number of data points.
        # So need to set whole number.
        # It'll be based on proportion given * number of data points in previous day.
        self.omicron = config['omicron'] * self.dataset_size

        # make sure this is done only after we set outlier deletion point! This is because we want the deletion point
        # to be based on "previous day dataset size"!
        self.dataset_size = input_dataset.shape[0]

        # make sure this is run after self.dataset_size is set to the current day's dataset size!
        self.mu = self.calculate_density_threshold()

        self.progres_bar_interval = len(input_dataset) * 0.01

    def calculate_pref_dim_variance_threshold(self):
        """
        Calculate the variance threshold used the determine whether a dimension is preferred by a cluster.
        This variable changes with time because an input dataset for a given point in time have to be normalised to [0,
        1] based on their collective values prior to clustering due to possible experimental error. For example,
        data for t=0 may lie in range [0,1], but for t=1, data may lie in range [0.1,0.88] due to an experimental error.
        In order to capture all the artifacts, we need to set different variance threshold for t=0 and t=1.

        Args:
            input_dataset (numpy.array): 2d array containing input dataset for a given point in time.

        Returns:
            Float: variance threshold.
        """

        variance_threshold = float(self.config['delta'])

        # Variance threshold given must be 0 - 1. If not we need to terminate
        if variance_threshold > 1 or variance_threshold < 0:
            sys.exit("Given delta ({}) is out of range. Must be within 0-1.".format(variance_threshold))

        return variance_threshold

    def calculate_density_threshold(self):
        """
        Calculate density threshold based on the size of input dataset and density_threshold_proportion config (what's
        the proportion of input dataset to be used as density threshold value).

        Args:
            input_dataset (numpy.array): 2d array containing input dataset for a given point in time.

        Returns:
            Float: density threshold value.
        """
        return float(self.config['mu']) * self.dataset_size

    def online_microcluster_maintenance(self, input_dataset, input_dataset_daystamp, reset_param=True):
        """
        Perform HDDStream online microcluster maintenance. In summary, it adds new points (the one in the
        input_dataset above) into either existing potential microcluster or new/existing outlier microcluster.
        See section 4.2 in paper[1].

        Args:
            input_dataset (numpy.array): 2d array containing input dataset for a given point in time.
            input_dataset_daystamp (int): timestamp of the input dataset in day i.e. day 1, day 2, etc.
            reset_param (bool, optional): True if need to recalculate parameters that are dependent on the dataset.
                False otherwise.

        Returns:
            None.
        """
        if reset_param:
            self._set_dataset_dependent_parameters(input_dataset)

        logger = self.logger

        logger.info(f"Setting up online phase for timepoint {input_dataset_daystamp} with following params:\n"
                         f"Pcore density threshold factor(beta) = {self.beta}\n"
                         f"Decay rate(lambda) = {self.lambbda}\n"
                         f"Radius threshold(epsilon) = {self.epsilon}\n"
                         f"Max projected dimensionality(pi) = {self.pi}\n"
                         f"Density threshold(mu) = {self.mu} = {self.mu}\n"
                         f"Variance threshold(delta) = {self.delta}\n"
                         f"K = {self.k}\n"
                         f"PreDeCon epsilon(upsilon) = {self.upsilon}\n"
                         f"Outlier deletion point(omicron) = {self.omicron}\n")

        # Check whether we need to decay the cluster. We only decay it if this dataset is not for the same day as the
        #  previous dataset processed by the online microcluster maintenance.
        if (self.last_data_timestamp - input_dataset_daystamp) != 0:
            logger.info("Decaying and downgrading microclusters")
            # The time difference is converted to days because we only decay as each day has passed between datasets.
            interval = input_dataset_daystamp - self.last_data_timestamp
            self._decay_clusters_weight(interval)

            self.downgrade_microclusters()

            # Because we have dataset for new day, we want to reset the clusters' current_data_points_weight_sum
            for pmc in self.pcore_MC:
                # Save memory. Don't store every points.
                pmc.reset_points()
            for omc in self.outlier_MC:
                # Save memory. Don't store every points.
                omc.reset_points()

        num_datapoints = input_dataset.shape[0]

        logger.info("Starting online microcluster maintenance for timepoint {}".format(input_dataset_daystamp))
        # progress bar widget
        progress_bar = TqdmToLogger(logger, level=logging.INFO)
        for row in tqdm(range(num_datapoints), file=progress_bar, mininterval=1):
            # You may find sometimes the progress line doesn't work well. In that case uncomment below.
            datapoint = input_dataset[row]

            # trial1 contains boolean that indicates whether the point has successfully been added to a potential
            # microcluster. See Figure 1 in paper[1].
            trial1 = self._add_to_pcore(datapoint, input_dataset_daystamp, row)
            trial2 = False

            if not trial1:
                # code will get here if the point cannot be added to any potential microcluster. In this case we'll
                # see if we can add it to an outlier microcluster
                trial2 = self._add_to_outlier(datapoint, input_dataset_daystamp, row)

            # No need to check if trial2 is none as it won't even get there if trial1 is true.
            if not trial1 and not trial2:
                # We create a new outlier cluster for the datapoint.
                self._create_new_outlier_cluster(datapoint, input_dataset_daystamp, row)

        logger.info("Finish online microcluster maintenance for timepoint {}".format(input_dataset_daystamp))
        logger.info("Online maintenance yield {} pcores and {} outlier".format(
            len(self.pcore_MC), len(self.outlier_MC)))

        self.last_data_timestamp = input_dataset_daystamp

        self.offline_clustering(input_dataset_daystamp)

    def _decay_clusters_weight(self, interval):
        """
        Reduce the weight of all microclusters. This is called when a new dataset for next day arrive.
        Note we decay both the pcore and the outlier.

        Args:
            interval (float): Time difference between last dataset processed by online cluster maintenance
                and new dataset.

        Returns:
            None.
        """
        pcore_MCs = self.pcore_MC
        outlier_MCs = self.outlier_MC

        for pcore in pcore_MCs:
            self.decay_a_cluster_weight(interval, pcore)

        for outlier_mc in outlier_MCs:
            self.decay_a_cluster_weight(interval, outlier_mc)

    def decay_a_cluster_weight(self, interval, microcluster):
        """
        Method to decay a microcluster's weight.
        There is no need to update the preferred dimension as its variance would have remained the same.
        Remember, decay is affecting CF1, CF2, and weight, and variance is calculated purely from them.
        So if all of them change, then the variance will remain the same.

        Args:
            interval (float): Time difference between last dataset processed by online cluster maintenance
                and new dataset.
            microcluster: Microcluster whose weight is to be decayed.

        Returns:
            None.
        """
        decay_factor = 2 ** (-self.lambbda * interval)
        microcluster.CF1 *= decay_factor
        microcluster.CF2 *= decay_factor
        microcluster.cumulative_weight *= decay_factor

    def _add_to_pcore(self, datapoint, datapoint_timestamp, datapoint_idx):
        """
        Add point (datapoint) to a pcore microcluster.
        Args:
            datapoint (numpy.array): A point represented as an array of values, each containing the point's value for a
                dimension.
            datapoint_timestamp (int): The timepoint the datapoint is meant for.
            datapoint_idx (int): The index (or so called id) of the data point.

        Returns:
            bool: False if addition failed i.e. some conditions are not met, True if addition was performed.
        """
        closest_distance = None
        closest_cluster_index = None

        # To save time rather than keep on referring to the self attributes.
        pcore_MCs = self.pcore_MC
        delta_squared = self.delta_squared
        k = self.k
        pi = self.pi
        epsilon_squared = self.epsilon_squared

        # calculate distances between datapoint and all pcore MCs.
        for index, pmc in enumerate(pcore_MCs):

            # In the Figure 2 paper[1] line 3-4,
            # we want to just temporarily add data point to each microcluster to
            # see if the datapoint can fit in it by checking the microcluster's pdim. We don't want to interfere the
            # original microcluster, so we clone it and pretend to add a point it.
            temp_pmc = pmc.get_copy_with_new_point(datapoint, delta_squared, k)

            pdim_temp_pmc = (np.array(temp_pmc.preferred_dimension_vector) != 1).sum()

            if pdim_temp_pmc <= pi:
                # Rather than keeping array of potential microclusters, we just calculate distance to it, and compare
                # to see if there has been one that was closer before. If there hasn't
                # then store this as closest. Otherwise leave it.
                distance = pmc.get_projected_dist_to_point(datapoint)
                if closest_distance is None or distance < closest_distance:
                    closest_distance = distance
                    closest_cluster_index = index

        if closest_distance is not None:
            # We got here when there exists a potential microcluster that can accomodate the point. We then check to
            # see if the potential microcluster can actually accomodate the point i.e. its radius will not blow out
            # beyond the radius threshold. See line 14-15 in Figure 2 paper[1].
            tmp_closest_cluster = pcore_MCs[closest_cluster_index].get_copy_with_new_point(datapoint, delta_squared, k)
            projected_radius_squared = tmp_closest_cluster.calculate_projected_radius_squared()

            if projected_radius_squared <= epsilon_squared:

                # TODO: should the update preferred dimension done by add new point?
                pcore_MCs[closest_cluster_index].add_new_point(datapoint, datapoint_timestamp, datapoint_idx)
                pcore_MCs[closest_cluster_index].update_preferred_dimensions(delta_squared, k)
                return True
        return False

    def _add_to_outlier(self, datapoint, datapoint_timestamp, datapoint_idx):
        """
        Add a datapoint to outlier microcluster. This can be improved by consolidating it with the add to pcore
        since it's so similar.
        It will also upgrade an outlier microcluster to pcore microcluster if the conditions for pcore microcluster
        are all met.
        Refer to section 4.2.2 in paper [1].

        Args:
            datapoint (numpy.array): A point represented as an array of values, each containing a point's value for a
                dimension.
            datapoint_timestamp (int): The timepoint the datapoint is meant for.
            datapoint_idx (int): The index (or so called id) of the data point.

        Returns:
            bool: False if addition failed i.e. some conditions are not met, True if addition was performed.
        """
        closest_distance = None
        closest_cluster_index = None

        outlier_MCs = self.outlier_MC
        delta_squared = self.delta_squared
        k = self.k
        epsilon_squared = self.epsilon_squared

        # Find closest outlier microcluster.
        for index, omc in enumerate(outlier_MCs):
            distance = omc.get_projected_dist_to_point(datapoint)
            if closest_distance is None or distance < closest_distance:
                closest_distance = distance
                closest_cluster_index = index

        # TODO merge this with pcore one as it is very similar
        if closest_distance is not None:
            # We got here when there exists an outlier microcluster that can accomodate the point. We then check to
            # see if the outlier microcluster can actually accomodate the point i.e. its radius will not blow out
            # beyond the radius threshold.

            tmp_outlier_mc = outlier_MCs[closest_cluster_index].get_copy_with_new_point(datapoint, delta_squared, k)
            projected_radius_squared = tmp_outlier_mc.calculate_projected_radius_squared()

            if projected_radius_squared <= epsilon_squared:
                # TODO: should the update preferred dimension done by add new point?
                outlier_MCs[closest_cluster_index].add_new_point(datapoint, datapoint_timestamp, datapoint_idx)
                outlier_MCs[closest_cluster_index].update_preferred_dimensions(delta_squared, k)

                # From here on, we then check whether the outlier microcluster can be upgraded to pcore microcluster.
                self._upgrade_outlier_microcluster(outlier_MCs[closest_cluster_index])

                return True
        return False

    def _upgrade_outlier_microcluster(self, outlier_mc):
        """
        Method to upgrade an outlier microcluster if its weight and preferred dimensionality of the cluster
        conditions are obeyed.
        See definition 6 in paper[1].
        There is no need to check the radius as it should have been checked before we add new point to the outlier
        microcluster! i.e. before we even got here.

        Args:
            outlier_mc (:obj:`Microcluster`): Outlier Microcluster to be upgraded.

        Returns:
            None.
        """
        beta = self.beta
        mu = self.mu
        pi = self.pi
        pcore_MCs = self.pcore_MC

        weight_threshold_obeyed = outlier_mc.cumulative_weight >= beta * mu
        pdim_threshold_obeyed = np.array(outlier_mc.preferred_dimension_vector > 1).sum() <= pi

        if weight_threshold_obeyed and pdim_threshold_obeyed:

            if outlier_mc.prev_pcore_id is None:
                # never been upgraded. Need new id
                id_to_assign = self.pcore_MC_last_id
                self.pcore_MC_last_id += 1
            else:
                id_to_assign = outlier_mc.prev_pcore_id

            outlier_mc.id = [id_to_assign]
            self.outlier_MC.remove(outlier_mc)
            pcore_MCs.append(outlier_mc)



    def _create_new_outlier_cluster(self, datapoint, creation_time, datapoint_idx):
        """
        Create a new outlier microcluster for a datapoint and add it to the outlier microcluster list.

        Args:
            datapoint (numpy.array): A point represented as an array of values, each containing a point's value for a
                dimension.
            datapoint_idx (int): The index (or so called id) of the data point.
            creation_time (int): Time when the cluster is created.

        Returns:
            None.
        """
        outlier_MCs = self.outlier_MC

        id_to_assign = self.outlier_MC_last_id
        outlier_mc_id = set()
        outlier_mc_id.add(id_to_assign)

        num_datapoints = len(datapoint)
        outlier_mc = Microcluster(cf1=np.zeros(num_datapoints), cf2=np.zeros(num_datapoints), id=outlier_mc_id,
                                  creation_time_in_hrs=creation_time)
        # TODO: should the update preferred dimension done by add new point?
        outlier_mc.add_new_point(datapoint, creation_time, datapoint_idx)
        outlier_mc.update_preferred_dimensions(self.delta_squared, self.k)
        outlier_mc.update_prev_outlier_id(id_to_assign)

        outlier_MCs.append(outlier_mc)
        self.outlier_MC_last_id += 1

    def offline_clustering(self, dataset_daystamp):
        """
        Perform offline clustering step in HDDStream to extract core microclusters.
        Args:
            dataset_daystamp (int): Which day is the dataset for This is used just for logging purposes.

        Returns:
            None.
        """

        datapoints = {}

        num_core = 0
        pcore_MCs = self.pcore_MC
        epsilon_squared = self.epsilon_squared
        mu = self.mu
        pi = self.pi
        logger = self.logger

        for cluster in pcore_MCs:

            # For offline clustering, the core status of each cluster is determined by the cluster itself
            # rather than by PreDeCon.
            cluster_id = next(iter(cluster.id))

            cluster_is_core = cluster.is_core(epsilon_squared, mu, pi)
            if cluster_is_core:
                num_core += 1

            datapoints[cluster_id] = PredeconMC(centroid=cluster.cluster_centroids,
                                        id=cluster_id, is_core_cluster=cluster_is_core,
                                        cluster_CF1=cluster.CF1, cluster_CF2=cluster.CF2,
                                        cluster_cumulative_weight=cluster.cumulative_weight)
        num_pcore = len(pcore_MCs) - num_core
        logger.info(f'Starting offline clustering with {num_core} core clusters and {num_pcore} pcore clusters.')

        predecon_offline = PreDeCon(datapoints=datapoints, dataset_dimensionality=self.dataset_dimensionality,
                                    epsilon=self.upsilon,
                                    delta=self.delta,
                                    lambbda=pi,
                                    mu=mu,
                                    k=self.k)
        predecon_offline.run()

        self.final_clusters = predecon_offline.clusters
        logger.info('Finish offline clustering for dataset with timepoint: {}'.format(dataset_daystamp))
        logger.info("Offline clustering yield {} clusters.".format(len(self.final_clusters)))

    def downgrade_microclusters(self):
        self._downgrade_potential_microclusters()
        self._downgrade_outlier_microclusters()

    def _downgrade_potential_microclusters(self):
        """
        Downgrade a potential microcluster if its weight and preferred dimensionality of the cluster conditions are
        no longer obeyed as in Definition 6 in paper[1].
        """

        pcore_MCs = self.pcore_MC
        outlier_MCs = self.outlier_MC
        beta = self.beta
        mu = self.mu
        pi = self.pi

        for potential_cluster in pcore_MCs:
            weight_threshold_obeyed = potential_cluster.cumulative_weight < beta * mu
            pdim_threshold_obeyed = np.array(potential_cluster.preferred_dimension_vector > 1).sum() > pi

            if weight_threshold_obeyed or pdim_threshold_obeyed:

                # potential_cluster.id = list(range(len(self.outlier_MC), len(self.outlier_MC) + 1))
                potential_cluster.id = [potential_cluster.prev_outlier_id]
                pcore_MCs.remove(potential_cluster)
                outlier_MCs.append(potential_cluster)

    def _downgrade_outlier_microclusters(self):
        """
        Delete outlier microcluster. See section 4.4 in paper [1].
        """
        outlier_MCs = self.outlier_MC
        omicron = self.omicron
        for outlier_cluster in outlier_MCs:

            if outlier_cluster.cumulative_weight <= omicron:
                outlier_MCs.remove(outlier_cluster)
                del outlier_cluster


class TqdmToLogger(io.StringIO):
    """
    This is for logging progress bar purposes only.
    Credit to the original owner of the code:
        https://stackoverflow.com/questions/14897756/python-progress-bar-through-logging-module
    Output stream for TQDM which will output to logger module instead of
    the StdOut.
    """
    logger = None
    level = None
    buf = ''

    def __init__(self,logger,level=None):
        super(TqdmToLogger, self).__init__()
        self.logger = logger
        self.level = level or logging.INFO

    def write(self,buf):
        self.buf = buf.strip('\r\n\t ')

    def flush(self):
        self.logger.log(self.level, self.buf)
